import matplotlib as mpl
import numpy as np
from itertools import tee, islice
from matplotlib import colors as mcolors
from uuid import uuid4

ECHARTS16 = [
    "#5470c6",
    "#91cc75",
    "#fac858",
    "#ee6666",
    "#9a60b4",
    "#73c0de",
    "#3ba272",
    "#fc8452",
    "#27727b",
    "#ea7ccc",
    "#d7504b",
    "#e87c25",
    "#b5c334",
    "#fe8463",
    "#26c0c0",
    "#f4e001",
]


def pairwise(iterable):
    """This is not available in itertools until 3.10"""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def grouper(iterable, n):
    """Collect data into non-overlapping fixed-length chunks or blocks"""
    args = [iter(iterable)] * n
    return zip(*args)


def batched(iterable, n):
    """Batch data into lists of length n. The last batch may be shorter."""
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := list(islice(it, n)):
        yield batch


# Copy from seaborn/utils.py
def relative_luminance(color):
    """Calculate the relative luminance of a color according to W3C standards
    Parameters
    ----------
    color : matplotlib color or sequence of matplotlib colors
        Hex code, rgb-tuple, or html color name.
    Returns
    -------
    luminance : float(s) between 0 and 1
    """
    rgb = mcolors.colorConverter.to_rgba_array(color)[:, :3]
    rgb = np.where(rgb <= 0.03928, rgb / 12.92, ((rgb + 0.055) / 1.055) ** 2.4)
    lum = rgb.dot([0.2126, 0.7152, 0.0722])
    try:
        return lum.item()
    except ValueError:
        return lum


def get_colormap(cmap):
    try:
        return mpl.colormaps.get_cmap(cmap)
    except AttributeError:
        try:
            return mpl.colormaps.get(cmap)
        except AttributeError:
            return mpl.cm.get_cmap(cmap)


def get_canvas_size_by_data(
    shape, width=None, height=None, scale=0.3, aspect=1, max_side=15
):
    h, w = shape
    no_w = width is None
    no_h = height is None
    # if user set both side, the aspect is ignored

    if no_h & no_w:
        width = w * scale
        height = h * scale * aspect
    elif no_h:
        # recompute scale
        scale = width / w
        height = h * scale * aspect
    elif no_w:
        scale = height / h
        width = w * scale / aspect
    size = np.array([width, height])
    if size.max() > max_side:
        size = size / size.max() * max_side
    width, height = size

    return width, height


def get_plot_name(name=None, side=None, chart=None):
    if name is None:
        return f"{chart}-{side}-{uuid4().hex}"
    else:
        return name


def _check_side(side):
    """Check user input the correct word"""
    options = ["top", "bottom", "left", "right"]
    if side not in options:
        raise ValueError(f"`side` must be one of {options}.")



import pandas as pd
from legendkit import cat_legend

from marsilea.plotter.base import StatsBase, RenderSpec


class Range(StatsBase):
    """Range plot

    The range plot shows the range between two categories.
    The input data should be a DataFrame with two columns.

    Parameters
    ----------
    data : array-like or DataFrame
        The input data.
    items : array-like, default: None
        The names of the items.
    marker : str, default: 'o'
        The marker style.
    markersize : float, default: 50
        The size of the marker.
    color1 : str, default: '#F75940'
        The color of the first range.
    color2 : str, default: '#3DC7BE'
        The color of the second range.
    edgecolor1 : str, default: 'black'
        The edge color of the first range.
    edgecolor2 : str, default: 'black'
        The edge color of the second range.
    edgewidth : float, default: 1
        The width of the edge.
    linecolor : str, default: 'black'
        The color of the line.
    linewidth : float, default: 1
        The width of the line.
    label : str, default: None
        The label of the plot.

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> import marsilea as ma
        >>> import numpy as np
        >>> data = np.random.rand(10, 2)
        >>> range_data = np.random.randint(1, 100, (10, 2))
        >>> h = ma.Heatmap(data)
        >>> h.add_left(ma.plotter.Range(range_data, items=["A", "B"]))
        >>> h.render()


    """

    def __init__(
        self,
        data,
        items=None,
        marker="o",
        markersize=50,
        color1="#F75940",
        color2="#3DC7BE",
        edgecolor1="black",
        edgecolor2="black",
        edgewidth=1,
        linecolor="black",
        linewidth=1,
        label=None,
    ):
        if isinstance(data, pd.DataFrame):
            if items is None:
                items = data.columns
            data = data.to_numpy()
        else:
            data = data
            if items is None:
                items = [f"Item {i}" for i in range(data.shape[1])]
        data = self.data_validator(data, target="2d")
        self.set_data(data.T)
        self.set_label(label)
        self.items = items
        self.marker = marker
        self.markersize = markersize
        self.color1 = color1
        self.color2 = color2
        self.edgecolor1 = edgecolor1
        self.edgecolor2 = edgecolor2
        self.edgewidth = edgewidth
        self.linecolor = linecolor
        self.linewidth = linewidth

    def render_ax(self, spec: RenderSpec):
        ax = spec.ax
        data = spec.data.T

        for ix, (r1, r2) in enumerate(data):
            ix += 0.5
            x, y = [ix, ix], [r1, r2]
            if self.is_flank:
                x, y = y, x
            ax.scatter(
                [x[0]],
                [y[0]],
                s=self.markersize,
                marker=self.marker,
                color=self.color1,
                edgecolor=self.edgecolor1,
                zorder=1,
            )
            ax.scatter(
                [x[1]],
                [y[1]],
                s=self.markersize,
                marker=self.marker,
                color=self.color2,
                edgecolor=self.edgecolor2,
                zorder=1,
            )
            ax.plot(x, y, color=self.linecolor, linewidth=self.linewidth, zorder=0)

        if self.is_flank:
            ax.set_ylim(0, len(data))
        else:
            ax.set_xlim(0, len(data))
        if self.side == "left":
            ax.invert_xaxis()

    def get_legends(self):
        return [
            cat_legend(
                colors=[self.color1, self.color2], labels=self.items, title=self.label
            )
        ]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Iterable, Sequence, Dict

import numpy as np
import pandas as pd
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.offsetbox import AnchoredText
from seaborn import despine

from .._deform import Deformation
from ..exceptions import DataError, SplitConflict


# class DataValidator:
#
#     @singledispatch(np.ndarray)
#     def parse(self, data):
#         pass


class DataLoader:
    """Handle user data"""

    def __init__(self, data, target="2d", plotter: str = None):
        self.plotter = plotter
        self.data = self.from_any(data, target=target)

    def get_array(self):
        if np.ma.isMaskedArray(self.data):
            return self.data
        return np.asarray(self.data)

    def from_any(self, data, target="2d"):
        if isinstance(data, pd.DataFrame):
            return self.from_dataframe(data, target=target)
        elif isinstance(data, np.ndarray):
            return self.from_ndarray(data, target=target)
        elif isinstance(data, Iterable):
            return self.from_iterable(data, target=target)
        else:
            raise TypeError(f"Your input data with type {type(data)} is not supported.")

    def from_dataframe(self, data: pd.DataFrame, target="2d"):
        ndata = data.to_numpy()
        if target == "1d":
            self._check_1d_compat(ndata)
            ndata = ndata.flatten()
        return ndata

    def from_ndarray(self, data: np.ndarray, target="2d"):
        ndata = data
        if target == "1d":
            self._check_1d_compat(ndata)
            ndata = ndata.flatten()
        if (target == "2d") & (ndata.ndim == 1):
            ndata = ndata.reshape(1, -1)
        return ndata

    def from_iterable(self, data: Iterable, target="2d"):
        try:
            ndata = np.asarray(data)
        except Exception:
            msg = f"Cannot handle your data with type {type(data)}"
            if self.plotter is not None:
                msg += f" for {self.plotter}."
            raise DataError(msg)
        return self.from_ndarray(ndata, target=target)

    def _check_1d_compat(self, data: np.ndarray):
        if data.ndim == 2:
            row, col = data.shape
            if not (row == 1) or (col == 1):
                if self.plotter is not None:
                    msg = f"{self.plotter} requires 1D data as input."
                else:
                    msg = "Require 1D data as input."
                raise DataError(msg)


@dataclass(repr=False)
class RenderSpec:
    """The container class that holds the \
    rendering data and parameters for each axes

    Attributes
    ----------
    data : Any
        The data to be rendered
    params : List[Dict]
        The parameters for each data
    group_data : Any
        The group data
    group_params : Any
        The group parameters
    current_ix : int
        The current index of the axes
    total : int
        The total number of axes


    """

    ax: Axes

    data: Any
    params: List[Dict] = None
    group_data: Any = None
    group_params: Any = None

    current_ix: int = 0

    total: int = 1

    def __repr__(self):
        return f"{self.__class__.__name__}(id={self.current_ix}, total={self.total})"

    @property
    def is_first(self):
        return self.current_ix == 0

    @property
    def is_last(self):
        return self.current_ix == self.total - 1


class RenderPlanLabel:
    label_props = {
        "left": dict(loc="center right", bbox_to_anchor=(0, 0.5)),
        "right": dict(loc="center left", bbox_to_anchor=(1, 0.5)),
        "top": dict(
            loc="lower center", bbox_to_anchor=(0.5, 1), prop=dict(rotation=90)
        ),
        "bottom": dict(
            loc="upper center", bbox_to_anchor=(0.5, 0), prop=dict(rotation=90)
        ),
    }

    label_loc = {
        "top": "right",
        "bottom": "right",
        "left": "bottom",
        "right": "bottom",
    }

    def __init__(self, label, loc=None, props=None):
        self.label = label
        self.loc = loc
        self.props = {} if props is None else props

    def add(self, axes, side):
        if side != "main":
            if isinstance(axes, Sequence):
                if self.loc in ["top", "left"]:
                    label_ax = axes[0]
                else:
                    label_ax = axes[-1]
            else:
                label_ax = axes

            if self.loc is None:
                loc = self.label_loc[side]
            else:
                loc = self.loc
            label_props = self.label_props[loc]
            bbox_loc = label_props["loc"]
            bbox_to_anchor = label_props["bbox_to_anchor"]
            prop = label_props.get("prop", {})
            if self.props is not None:
                prop.update(self.props)

            title = AnchoredText(
                self.label,
                loc=bbox_loc,
                bbox_to_anchor=bbox_to_anchor,
                prop=prop,
                pad=0.3,
                borderpad=0,
                bbox_transform=label_ax.transAxes,
                frameon=False,
            )
            label_ax.add_artist(title)


class MetaRenderPlan(type):
    """Metaclass for RenderPlan"""

    def __init__(cls, name, bases, attrs):
        allow_labeling = attrs.get("allow_labeling", False)
        if allow_labeling:

            def new_render(self, axes):
                attrs["render"](self, axes)
                self._plan_label.add(axes, self.side)

            setattr(cls, "render", new_render)


class RenderPlan:
    """The base class for every plotter in Marsilea

    Attributes
    ----------
    name : str
        The name of the plot,
        can be used to retrieve the render axes and corresponding legend.
    side : str, default: 'top'
        Which side to render this plot
    size : float
    allow_split : bool, default: True
        Use to mark if the RenderPlan can be split
    render_main : bool, default: False
        Use to mark if the RenderPlan can be rendered on main canvas
    zorder : int, default: 0
        This only works if the RenderPlan is rendered on main canvas

    """

    name: str = None
    size: float = None
    side: str = "top"
    # label if a render plan
    # can be used on split axes
    allow_split: bool = True
    zorder: int = 0
    deform: Deformation = None
    # If True, this RenderPlan can be rendered on main ax
    render_main = False
    label: str = ""
    allow_labeling: bool = True

    _data: Sequence[np.ndarray] = None
    _is_datasets: bool = False
    _params: List[Dict[str, Any]] = None
    _group_data: List = None
    _group_params: List[Dict[str, Any]] = None
    _plan_label: RenderPlanLabel = None
    _split_regroup: Sequence[float] = None

    def __repr__(self):
        side_str = f"side='{self.side}'"
        zorder_str = f"zorder={self.zorder}"

        if self.name is None:
            chunks = [side_str, zorder_str]
        else:
            chunks = [f"name='{self.name}'", side_str, zorder_str]
        return f"{self.__class__.__name__}" f"({', '.join(chunks)})"

    def set(self, **kwargs):
        for k, v in kwargs.items():
            if k == "side":
                self.set_side(v)
            else:
                self.__setattr__(k, v)

    def set_side(self, side):
        self.side = side

    def set_size(self, size):
        self.size = size

    def set_label(self, label, loc=None, props=None):
        self.label = label
        self._plan_label = RenderPlanLabel(label, loc=loc, props=props)

    def get_deform_func(self):
        if self.has_deform:
            if self.side == "main":
                return self.deform.transform
            elif self.is_flank:
                return self.deform.transform_row
            else:
                return self.deform.transform_col

    def reindex_by_chunk(self, group_data):
        if group_data is not None:
            if self.has_deform:
                group_data = np.asarray(group_data)
                if (
                    self.deform.is_col_split
                    & (self.is_body | (self.side == "main"))
                    & self.deform.is_col_cluster
                ):
                    return group_data[self.deform.col_chunk_index]
                elif (
                    self.deform.is_row_split
                    & (self.is_flank | (self.side == "main"))
                    & self.deform.is_row_cluster
                ):
                    return group_data[self.deform.row_chunk_index]
            return group_data

    def set_deform(self, deform: Deformation):
        self.deform = deform

    def get_data(self):
        return self._data

    def set_data(self, *data):
        self._data = data

    def get_params(self):
        return self._params

    def set_params(self, params: Dict[str, List]):
        params_names = params.keys()
        spec_params = []
        for d in zip(*params.values()):
            spec_params.append(dict(zip(params_names, d)))
        self._params = spec_params

    def set_group_data(self, group_data):
        self._group_data = group_data

    def get_group_data(self):
        return self._group_data

    def set_group_params(self, group_params: Dict[str, List]):
        params_names = group_params.keys()
        spec_params = []
        for d in zip(*group_params.values()):
            spec_params.append(dict(zip(params_names, d)))
        self._group_params = spec_params

    def get_group_params(self):
        return self._group_params

    def data_validator(self, data, target="2d"):
        name = self.__class__.__name__
        loader = DataLoader(data, target=target, plotter=name)
        return loader.get_array()

    # def get_render_data(self):
    #
    #     data = self.get_data()
    #
    #     if self.has_deform:
    #         self.get_deform_func()
    #
    #     else:
    #         if len(data) == 1:
    #             return data[0]
    #         else:
    #             return data

    def _get_split_render_spec(self, axes):
        datasets = self.get_data()
        params = self.get_params()
        group_params = self.get_group_params()
        deform_func = self.get_deform_func()
        if len(datasets) == 1:
            spec_data = deform_func(datasets[0])
        else:
            datasets = [deform_func(d) for d in datasets]
            spec_data = [d for d in zip(*datasets)]

        n = len(spec_data)
        if params is not None:
            params = deform_func(np.asarray(params, dtype=object))
        else:
            params = [None for _ in range(n)]

        if group_params is not None:
            group_params = self.reindex_by_chunk(group_params)
        else:
            group_params = [None for _ in range(n)]

        spec_list = []
        total = len(axes)
        dispatch = zip(axes, spec_data, params, group_params)
        for i, (ax, d, p, gp) in enumerate(dispatch):
            spec = RenderSpec(
                ax=ax, data=d, params=p, group_params=gp, current_ix=i, total=total
            )
            spec_list.append(spec)
        return spec_list

    def _get_intact_render_spec(self, ax):
        datasets = self.get_data()
        params = self.get_params()

        if self.has_deform:
            deform_func = self.get_deform_func()
            if datasets is None:
                spec_data = None
            elif len(datasets) == 1:
                spec_data = deform_func(datasets[0])
            else:
                spec_data = [deform_func(d) for d in datasets]
            if params is not None:
                params = deform_func(np.asarray(params, dtype=object))
        else:
            spec_data = datasets[0] if len(datasets) == 1 else datasets
        return RenderSpec(ax=ax, data=spec_data, params=params)

    def get_render_spec(self, axes):
        try:
            if self.is_split:
                return self._get_split_render_spec(axes)
            else:
                return self._get_intact_render_spec(axes)
        except Exception as _:
            raise DataError(
                f"Please check your data input "
                f"with {self.__class__.__name__} at '{self.side}'"
            )

    @property
    def has_deform(self):
        """If the RenderPlan has Deformation"""
        return self.deform is not None

    @property
    def is_body(self):
        """Draw on top and bottom"""
        return self.side in ["top", "bottom"]

    @property
    def is_flank(self):
        """Draw on left and right"""
        return self.side in ["right", "left"]

    @property
    def is_split(self):
        """
        For the RenderPlan to self-aware of whether its render canvas
        will be split. Useful to determine how to get render data.
        """
        if self.deform is not None:
            if self.is_body & self.deform.is_col_split:
                return True
            if self.is_flank & self.deform.is_row_split:
                return True
            if (self.side == "main") & self.deform.is_split:
                return True
        return False

    def render_ax(self, spec: RenderSpec):
        """The major rendering function
        Define how the plot is drawn
        """
        pass

    def render(self, axes):
        """
        This function will be call when render a plot

        - If the canvas is split, :meth:`render_axes` will be called.
        - If only one ax is passed, :meth:`render_ax` will be called.

        .. code-block:: python

            if self.is_split:
                self.render_axes(axes)
            else:
                self.render_ax(axes, self.get_render_data())

        """
        if self.is_split:
            for spec in self.get_render_spec(axes):
                self.render_ax(spec)
        else:
            self.render_ax(self.get_render_spec(axes))

        if self.allow_labeling:
            if self._plan_label is not None:
                self._plan_label.add(axes, self.side)

    def get_canvas_size(
        self, figure, main_height=None, main_width=None, **kwargs
    ) -> float:
        """
        If the size is unknown before rendering, this function must be
        implemented to return the canvas size in inches.
        """
        return self.size

    def get_legends(self) -> List[Artist] | None:
        """
        This should define the legend return by the RenderPlan.

        The return object could be any :class:`matplotlib.artist.Artist`

        .. note::
            :class:`matplotlib.colorbar.Colorbar` is not an :class:`Artist`.
            Use :class:`legendkit.colorart`

        """
        return None

    def set_legends(self, *args, **kwargs):
        raise NotImplementedError

    def update_main_canvas_size(self):
        pass

    def set_split_regroup(self, ratio):
        self._split_regroup = ratio

    def get_split_regroup(self):
        if self._split_regroup is None:
            return None
        if self.is_flank:
            return self._split_regroup[::-1]
        return self._split_regroup


class StatsBase(RenderPlan):
    """A base class for rendering statistics plot"""

    render_main = True
    orient = None
    axis_options = None

    def get_orient(self):
        if self.orient is None:
            return "h" if self.is_flank else "v"
        return self.orient

    def get_deform_func(self):
        if self.has_deform:
            orient = self.get_orient()
            if self.side == "main":
                orient_mapper = {"h": "horizontally", "v": "vertically"}

                if ((orient == "v") & self.deform.is_row_split) or (
                    (orient == "h") & self.deform.is_col_split
                ):
                    plot_dir = orient_mapper[self.get_orient()]
                    msg = (
                        f"{self.__class__.__name__} is oriented "
                        f"{plot_dir} should only be split {plot_dir}"
                    )
                    raise SplitConflict(msg)

            if self.get_orient() == "v":
                return self.deform.transform_col
            else:
                return self.deform.transform_row

    def _setup_axis(self, ax):
        if self.get_orient() == "h":
            despine(ax=ax, left=True)
            ax.tick_params(left=False, labelleft=False, bottom=True, labelbottom=True)
        else:
            despine(ax=ax, bottom=True)
            ax.tick_params(left=True, labelleft=True, bottom=False, labelbottom=False)

    def align_lim(self, axes):
        is_inverted = False
        lim_low = []
        lim_up = []

        is_h = self.get_orient() == "h"
        for ax in axes:
            if is_h:
                low, up = ax.get_xlim()
                if ax.xaxis_inverted():
                    is_inverted = True
                    low, up = up, low
            else:
                low, up = ax.get_ylim()
                if ax.yaxis_inverted():
                    is_inverted = True
                    low, up = up, low
            lim_up.append(up)
            lim_low.append(low)
        lims = [np.min(lim_low), np.max(lim_up)]
        if is_inverted:
            lims = lims[::-1]
        for ax in axes:
            ax.set_xlim(*lims) if is_h else ax.set_ylim(*lims)

    def render(self, axes):
        # check if self._plan_label.loc is None
        is_loc = False
        props = None
        if self._plan_label is not None:
            if self._plan_label.loc is not None:
                is_loc = True
            props = self._plan_label.props
        if self.is_split:
            for spec in self.get_render_spec(axes):
                self.render_ax(spec)
            self.align_lim(axes)

            for i, ax in enumerate(axes):
                # leave axis for the first ax
                if (i == 0) & (self.get_orient() == "v"):
                    self._setup_axis(ax)
                    if not is_loc:
                        ax.set_ylabel(self.label, fontdict=props)
                # leave axis for the last ax
                elif (i == len(axes) - 1) & (self.get_orient() == "h"):
                    self._setup_axis(ax)
                    if not is_loc:
                        ax.set_xlabel(self.label, fontdict=props)
                else:
                    ax.set_axis_off()
        else:
            # axes.set_axis_off()
            self.render_ax(self.get_render_spec(axes))
            self._setup_axis(axes)
            if self.label is not None:
                if self.get_orient() == "v":
                    if not is_loc:
                        axes.set_ylabel(self.label, fontdict=props)
                else:
                    if not is_loc:
                        axes.set_xlabel(self.label, fontdict=props)

        if self.allow_labeling & is_loc:
            self._plan_label.add(axes, self.side)



__all__ = [
    "Bar",
    "Box",
    "Boxen",
    "Violin",
    "Point",
    "Strip",
    "Swarm",
    "Arc",
    "Numbers",
    "StackBar",
    "CenterBar",
    "RenderPlan",
    "SeqLogo",
    "Colors",
    "ColorMesh",
    "SizedMesh",
    "MarkerMesh",
    "TextMesh",
    "Labels",
    "AnnoLabels",
    "Title",
    "Chunk",
    "FixedChunk",
    "Area",
    "Image",
    "Emoji",
    "Range",
]

from ._seaborn import Bar, Box, Boxen, Violin, Point, Strip, Swarm
from .arc import Arc
from .bar import Numbers, StackBar, CenterBar
from .base import RenderPlan
from .bio import SeqLogo
from .mesh import Colors, ColorMesh, SizedMesh, MarkerMesh, TextMesh
from .text import Labels, AnnoLabels, Title, Chunk, FixedChunk

from .images import Emoji, Image
from .area import Area
from .range import Range



# The implementation of Image in matplotlib may suffer from compatibility
# issues across different rendering backend at different DPI. Currently
# not a public API.
from functools import partial
from hashlib import sha256
from numbers import Number
from pathlib import Path

import numpy as np
from PIL import Image as PILImage
from matplotlib.image import imread, BboxImage
from matplotlib.transforms import Bbox
from platformdirs import user_cache_dir

from .base import RenderPlan


def _cache_remote(url, cache=True):
    try:
        import requests
    except ImportError:
        raise ImportError("Required requests, try `pip install requests`.")
    data_dir = Path(user_cache_dir(appname="Marsilea"))
    data_dir.mkdir(exist_ok=True, parents=True)

    hasher = sha256()
    hasher.update(url.encode("utf-8"))
    fname = hasher.hexdigest()

    dest = data_dir / fname
    if not (cache and dest.exists()):
        r = requests.get(url, headers={"User-Agent": "Mozilla/5.0"})
        r.raise_for_status()
        with open(dest, "wb") as f:
            f.write(r.content)

    return dest


class Image(RenderPlan):
    """Plot static images

    Parameters
    ----------
    images : array of image
        You can input either path to the image, URL, or numpy array.
    align : {"center", "top", "bottom", "left", "right"}, default: "center"
        The alignment of the images.
    scale : float, default: 1
        The scale of the images.
    spacing : float, default: 0.1
        The spacing between images, a value between 0 and 1,
        relative to the image container size.
    resize : int or tuple, default: None
        The size to resize the images.

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> import marsilea as ma
        >>> c = ma.ZeroWidth(height=2)
        >>> c.add_right(
        ...     ma.plotter.Image(
        ...         [
        ...             "https://www.iconfinder.com/icons/4375050/download/png/512",
        ...             "https://www.iconfinder.com/icons/8666426/download/png/512",
        ...             "https://www.iconfinder.com/icons/652581/download/png/512",
        ...         ],
        ...         align="right",
        ...     ),
        ...     pad=0.1,
        ... )
        >>> c.add_right(
        ...     ma.plotter.Labels(["Python", "Rust", "JavaScript"], fontsize=20), pad=0.1
        ... )
        >>> c.render()

    """

    def __init__(
        self,
        images,
        align="center",
        scale=1,
        spacing=0.1,
        resize=None,
    ):
        self.images_mapper = {}

        for i, img in enumerate(images):
            if isinstance(img, str):
                # Read from URL
                if img.startswith("http") or img.startswith("https"):
                    img = imread(_cache_remote(img))
                else:
                    # Read from string path
                    img = imread(img)
            # Read from Path
            elif isinstance(img, Path):
                img = imread(img)
            else:
                # Read from array interface
                img = np.asarray(img)

            self.images_mapper[i] = img

        self.images_codes = np.asarray(list(self.images_mapper.keys()))

        self.images = images
        self.align = align
        self.scale = scale
        if not 0 <= spacing <= 1:
            raise ValueError("spacing should be between 0 and 1")
        self.spacing = spacing
        if resize is not None:
            if isinstance(resize, Number):
                resize = (int(resize), int(resize))
            for i, img in self.images_mapper.items():
                self.images_mapper[i] = img.resize(resize, PILImage.Resampling.LANCZOS)
        for i, img in self.images_mapper.items():
            self.images_mapper[i] = np.asarray(img)
        self.set_data(self.images_codes)

    def _get_bbox_imges(
        self, ax, imgs, scale=1, align="center", ax_height=None, ax_width=None
    ):
        locs = np.linspace(0, 1, len(imgs) + 1)
        slot_size = locs[1] - locs[0]
        locs = locs[:-1] + slot_size * self.spacing / 2

        xmin, ymin = ax.transAxes.transform((0, 0))
        xmax, ymax = ax.transAxes.transform((1, 1))

        if ax_width is None or ax_width == 0:
            ax_width = xmax - xmin
        if ax_height is None or ax_height == 0:
            ax_height = ymax - ymin

        base_dpi = ax.get_figure().get_dpi()

        bbox_images = []
        imgaes_sizes = []

        if self.is_body:
            fit_width = ax_width / len(imgs) * (1 - self.spacing)
            for loc, img in zip(locs, imgs):
                height, width = img.shape[:2]
                fit_height = height / width * fit_width

                fit_scale_width = fit_width * scale
                fit_scale_height = fit_height * scale

                offset = (fit_width - fit_scale_width) / 2 / ax_width
                loc += offset

                if align == "top":
                    loc_y = 1 - fit_scale_height / ax_height
                elif align == "bottom":
                    loc_y = 0
                else:
                    loc_y = 0.5 - fit_scale_height / 2 / ax_height

                def img_bbox(renderer, loc, loc_y, width, height, base_dpi):
                    factor = renderer.dpi / base_dpi
                    x0, y0 = ax.transData.transform((loc, loc_y))
                    return Bbox.from_bounds(x0, y0, width * factor, height * factor)

                memorized_img_bbox = partial(
                    img_bbox,
                    loc=loc,
                    loc_y=loc_y,
                    width=fit_scale_width,
                    height=fit_scale_height,
                    base_dpi=base_dpi,
                )

                i1 = BboxImage(memorized_img_bbox, data=img)
                bbox_images.append(i1)
                imgaes_sizes.append(fit_scale_height)
        else:
            fit_height = ax_height / len(imgs) * (1 - self.spacing)
            for loc, img in zip(locs, imgs[::-1]):
                height, width = img.shape[:2]
                fit_width = width / height * fit_height

                fit_scale_width = fit_width * scale
                fit_scale_height = fit_height * scale

                offset = (fit_height - fit_scale_height) / 2 / ax_height
                loc += offset

                if align == "right":
                    loc_x = 1 - fit_scale_width / ax_width
                elif align == "left":
                    loc_x = 0
                else:
                    loc_x = 0.5 - fit_scale_width / 2 / ax_width

                def img_bbox(renderer, loc_x, loc, width, height, base_dpi):
                    factor = renderer.dpi / base_dpi
                    x0, y0 = ax.transData.transform((loc_x, loc))
                    return Bbox.from_bounds(x0, y0, width * factor, height * factor)

                memorized_img_bbox = partial(
                    img_bbox,
                    loc_x=loc_x,
                    loc=loc,
                    width=fit_scale_width,
                    height=fit_scale_height,
                    base_dpi=base_dpi,
                )

                i1 = BboxImage(memorized_img_bbox, data=img)
                bbox_images.append(i1)
                imgaes_sizes.append(fit_scale_width)

        return bbox_images, max(imgaes_sizes)

    def render_ax(self, spec):
        ax = spec.ax
        imgs = [self.images_mapper[i] for i in spec.data]
        bbox_images, _ = self._get_bbox_imges(ax, imgs)
        for i in bbox_images:
            ax.add_artist(i)
        ax.set_axis_off()

    def get_canvas_size(
        self, figure, main_height=None, main_width=None, **kwargs
    ) -> float:
        ax = figure.add_subplot(111)
        imgs = [self.images_mapper[i] for i in self.images_codes]
        _, size = self._get_bbox_imges(
            ax, imgs, ax_width=main_width, ax_height=main_height
        )
        ax.remove()
        return size


# ======== EMOJI Plotter ========

TWEMOJI_CDN = "https://cdn.jsdelivr.net/gh/twitter/twemoji/assets/72x72/"


class Emoji(Image):
    """Have fun with emoji images

    The emoji images are from `twemoji <https://twemoji.twitter.com/>`_.

    You can will all twemoji from `here <https://twemoji-cheatsheet.vercel.app/>`_

    Parameters
    ----------
    codes : array of str
        The emoji codes. You can input either unicode or short code.
    lang : str, default: "en"
        The language of the emoji.
    scale : float, default: 1
        The scale of the emoji.
    spacing : float, default: 0.1
        The spacing between emoji, a value between 0 and 1,
        relative to the emoji container size.

    Examples
    --------

    .. plot::

        >>> import marsilea as ma
        >>> c = ma.ZeroHeight(width=2)
        >>> c.add_top(ma.plotter.Emoji("😆😆🤣😂😉😇🐍🦀🦄"))
        >>> c.render()

    """

    def __init__(self, codes, lang="en", scale=1, spacing=0.1, **kwargs):
        try:
            import emoji
        except ImportError:
            raise ImportError("Required emoji, try `pip install emoji`.")

        urls = []
        for i in codes:
            i = emoji.emojize(i, language=lang)
            if not emoji.is_emoji(i):
                raise ValueError(f"{i} is not a valid emoji")
            c = f"{ord(i):X}".lower()
            urls.append(f"{TWEMOJI_CDN}{c}.png")

        super().__init__(urls, scale=scale, spacing=spacing, **kwargs)



# The API design of upsetplot is highly inspired from UpsetPlot
# https://github.com/jnothman/UpSetPlot

from __future__ import annotations

from collections import Counter

import numpy as np
import pandas as pd
from itertools import cycle
from legendkit import ListLegend
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle, Patch
from typing import List, Set, Mapping

from .base import WhiteBoard
from .plotter import (
    Numbers,
    Labels,
    StackBar,
    Bar,
    Box,
    Boxen,
    Violin,
    Point,
    Strip,
    Swarm,
)
from .utils import get_canvas_size_by_data


def _get_sets_table(binary_table):
    cardinality = binary_table.groupby(list(binary_table.columns), observed=True).size()
    sets_table = pd.DataFrame(cardinality, columns=["cardinality"])
    sets_table["degree"] = sets_table.index.to_frame().sum(axis=1)
    return sets_table


class UpsetData:
    """Handle multiple sets

    Normally, the construction methods are used to create a UpsetData

    Terminology that you might not be familiar with:
        - set: a collection of unique items
        - subset: intersection of sets
        - degree: The number of sets that intersect with each other
        - cardinality: The number of items in the subset

    Parameters
    ----------
    data : bool matrix
        A one-hot encode matrix indicates if an item is in a set.
        Columns are sets and rows are items
    sets_names : optional, array
        The name of sets
    items : optional, array
        The name of items
    sets_attrs : optional, pd.DataFrame
        The attributes of sets, the input index should be the
        same as names
    items_attrs : optional, pd.DataFrame
        The attributes of items, the input index should be the
        same as items

    Examples
    --------

    .. plot::
        :context: close-figs

        from marsilea.upset import UpsetData
        sets = [[1,2,3,4], [3,4,5,6]]
        data = UpsetData.from_sets(sets)


    """

    def __repr__(self):
        nitems, nsets = self._binary_table.shape
        return f"UpsetData: {nsets} sets, {nitems} items"

    def __init__(
        self, data, sets_names=None, items=None, sets_attrs=None, items_attrs=None
    ):
        if isinstance(data, pd.DataFrame):
            if sets_names is None:
                sets_names = data.columns.tolist()
            if items is None:
                items = data.index.tolist()
            data = data.to_numpy()
        if sets_names is None:
            raise ValueError("The name of sets must be provided")
        if items is None:
            raise ValueError("The name of items must be provided")

        assert len(sets_names) == len(set(sets_names)), "Duplicates in set names"
        assert len(items) == len(set(items)), "Duplicates in items"

        if sets_attrs is not None:
            sets_attrs = sets_attrs.loc[list(sets_names)]
        self._sets_attrs = sets_attrs

        if items_attrs is not None:
            items_attrs = items_attrs.loc[list(items)]
        self._items_attrs = items_attrs

        self._binary_table = pd.DataFrame(columns=sets_names, index=items, data=data)
        self._sets_table = _get_sets_table(self._binary_table)

    def filter(
        self,
        min_degree=None,
        max_degree=None,
        min_cardinality=None,
        max_cardinality=None,
    ):
        """Filter by degree or cardinality

        Parameters
        ----------
        min_degree : int
            The minimum degree
        max_degree : int
            The maximum degree
        min_cardinality : int
            The minimum cardinality
        max_cardinality : int
            The maximum cardinality

        """
        sets_table = self._sets_table
        if min_degree is not None:
            sets_table = sets_table[sets_table["degree"] >= min_degree]
        if max_degree is not None:
            sets_table = sets_table[sets_table["degree"] <= max_degree]
        if min_cardinality is not None:
            sets_table = sets_table[sets_table["cardinality"] >= min_cardinality]
        if max_cardinality is not None:
            sets_table = sets_table[sets_table["cardinality"] <= max_cardinality]
        self._sets_table = sets_table
        return self

    def sort_subsets(self, by="degree", ascending=False):
        """Sort the subsets by degree or cardinality

        Parameters
        ----------
        by : str
            Sort by either `degree` or `cardinality`
        ascending : bool
            Sort in ascending order if True


        """
        if by not in ["degree", "cardinality"]:
            raise ValueError("Sort by either `degree` or `cardinality`")
        if by == "cardinality":
            self._sets_table.sort_values(
                by=by, ascending=not ascending, inplace=True, kind="stable"
            )
        else:
            matrix = self._sets_table.index.to_frame().reset_index(drop=True)
            _, num = matrix.shape

            matrix["SUM"] = matrix.sum(axis=1)

            reorder_ix = []
            for n, df in matrix.groupby("SUM"):
                del df["SUM"]
                rows = df.to_numpy()
                c_rows = []
                for row in rows:
                    str_digit = row[::-1].astype(str)
                    str_row = "".join(str_digit)
                    c_rows.append(int(str_row))
                ix = np.argsort(c_rows)
                rix = df.index[ix].to_list()
                if not ascending:
                    rix = rix[::-1]
                reorder_ix += rix

            if ascending:
                reorder_ix = reorder_ix[::-1]
            self._sets_table = self._sets_table.iloc[reorder_ix, :]

        return self

    def sort_sets(self, ascending=False, order=None):
        """Control the order of sets

        Parameters
        ----------
        ascending : bool
            Sort in ascending order if True
        order : list
            Explicitly specify the order of sets

        """
        if order is not None:
            sets_names = order
        else:
            sets_sizes = self.sets_size().sort_values(
                ascending=ascending, kind="stable"
            )
            sets_names = sets_sizes.index.to_list()
        self._binary_table = self._binary_table.loc[:, sets_names]
        self._sets_table = self._sets_table.reorder_levels(order=sets_names)
        if self._sets_attrs is not None:
            self._sets_attrs = self._sets_attrs.loc[sets_names]
        return self

    def mark(
        self,
        present=None,
        absent=None,
        min_cardinality=None,
        max_cardinality=None,
        min_degree=None,
        max_degree=None,
    ):
        sets_table = self._sets_table
        marks = np.ones(len(sets_table), dtype=int)

        if isinstance(present, str):
            present = [present]
        if isinstance(absent, str):
            absent = [absent]
        if present is not None:
            for cat in present:
                marks = marks & (sets_table.index.get_level_values(cat) == 1)
        if absent is not None:
            for cat in absent:
                marks = marks & (sets_table.index.get_level_values(cat) == 0)

        if min_cardinality is not None:
            marks = marks & (sets_table["cardinality"] >= min_cardinality)
        if max_cardinality is not None:
            marks = marks & (sets_table["cardinality"] <= max_cardinality)
        if min_degree is not None:
            marks = marks & (sets_table["degree"] >= min_degree)
        if min_degree is not None:
            marks = marks & (sets_table["degree"] >= max_degree)
        return marks

    def reset(self):
        self._sets_table = _get_sets_table(self._binary_table)
        return self

    @classmethod
    def from_sets(
        cls,
        sets: List[Set],
        sets_names=None,
        sets_attrs: pd.DataFrame = None,
        items_attrs: pd.DataFrame = None,
    ) -> UpsetData:
        """Create UpsetData from a series of sets

        Parameters
        ----------
        sets : array of sets, dict
            The sets data
        sets_names : optional
            The name of sets, if name is not provided, it will
            be automatically named as "Set 1, Set 2, ..."
        sets_attrs : optional, pd.DataFrame
            The attributes of sets, the input index should be the
            same as sets_names
        items_attrs : optional, pd.DataFrame
            The attributes of items, the input index should be the
            same as items


        """
        items = set()
        new_sets = []
        new_names = []
        sets_total = {}

        if isinstance(sets, Mapping):
            it = sets.items()
        else:
            if sets_names is None:
                sets_names = [f"Set {i + 1}" for i in range(len(sets))]
            it = zip(sets_names, sets)

        for name, s in it:
            s = set(s)
            new_sets.append(s)
            new_names.append(name)
            items.update(s)
            sets_total[name] = len(s)

        items = np.array(list(items))
        data = []
        for name, s in zip(new_names, new_sets):
            d = [i in s for i in items]
            data.append(d)
        data = np.array(data, dtype=int).T
        container = cls(
            data,
            sets_names=new_names,
            items=items,
            sets_attrs=sets_attrs,
            items_attrs=items_attrs,
        )
        return container

    @classmethod
    def from_memberships(
        cls,
        items,
        items_names=None,
        sets_attrs: pd.DataFrame = None,
        items_attrs: pd.DataFrame = None,
    ):
        """Describe the sets an item are in

        Parameters
        ----------
        items : array of array of sets_names, dict
            The data of items
        items_names : optional
            The name of items, if name is not provided, it will
            be automatically named as "Item 1, Item 2, ..."
        sets_attrs : optional, pd.DataFrame
            The attributes of sets, the input index should be the
            same as sets_names
        items_attrs : optional, pd.DataFrame
            The attributes of items, the input index should be the
            same as items

        """

        sets = []
        new_items_names = []

        if isinstance(items, Mapping):
            for name, ss in items.items():
                new_items_names.append(name)
                sets.append({s: True for s in ss})
        else:
            for i, ss in enumerate(items):
                new_items_names.append(f"Item {i + 1}")
                sets.append({s: True for s in ss})

        if items_names is not None:
            new_items_names = items_names

        df = pd.DataFrame(sets).astype(float).fillna(False).astype(int)
        container = cls(
            df.to_numpy(),
            sets_names=df.columns,
            items=new_items_names,
            sets_attrs=sets_attrs,
            items_attrs=items_attrs,
        )
        return container

    def has_item(self, item):
        """Return a list of sets' name the item is in"""
        item_data = self._binary_table.loc[item]
        return item_data.loc[item_data == 1].index.tolist()

    def intersection(self, sets_name):
        """Return the items that are shared in different sets"""
        expr = "&".join([f"(`{s}`==1)" for s in sets_name])
        return self._binary_table.query(expr).index.tolist()

    def intersection_count(self):
        """The item has occurred in how many sets"""
        return self._binary_table.sum(axis=1)

    def binary_table(self):
        return self._binary_table

    def sets_table(self) -> pd.DataFrame:
        return self._sets_table

    def cardinality(self):
        """The number of items in intersections"""
        return self._sets_table["cardinality"]

    def degree(self):
        """Intersection between how many sets"""
        return self._sets_table["degree"]

    def sets_size(self):
        return self._binary_table.sum()[self.sets_names]

    @property
    def sets_names(self):
        return self._sets_table.index.names

    @property
    def items(self):
        return self._binary_table.index

    @property
    def sets_attrs(self):
        return self._sets_attrs

    @property
    def items_attrs(self):
        return self._items_attrs

    def get_items_attr(self, attr):
        """Return the attribute of items in the order of plotting

        Parameters
        ----------
        attr : key in items_attrs
            Retrieve the attribute of items

        Returns
        -------
            array of attribute

        """
        items_attrs = self._items_attrs
        sets_names = np.asarray(self.sets_names)

        data_collector = []

        for ix, row in self.sets_table().iterrows():
            s = sets_names[np.array(ix).astype(bool)]
            items = self.intersection(s)
            attr_data = items_attrs.loc[items][attr]
            data_collector.append(attr_data)
        return data_collector


class Upset(WhiteBoard):
    """Upset Plot

    Parameters
    ----------
    data : :class:`UpsetData`
        Upset data
    orient : str
        The orientation of the Upset plot
    sets_order : array of str
        The order of sets
    sets_color : array of color
        The color for each set, this will also change the bar color
        in set size
    sort_sets : {'ascending', 'descending'}
        The sorting order of sets
    sort_subsets : {'cardinality', 'degree'}
        How to sort the subset, 'size' will sort by intersection size,
        'degree' will sort by the number of intersected sets.
    min_degree, max_degree : int
        Select a fraction of subset to render by degree
    min_cardinality, max_cardinality : int
        Select a fraction of subset to render by cardinality
    color : color
        The main color to use
    shading : float
        The value to dilute the main color
    radius : float
        The size of the dot
    linewidth : float
        The width of lines that connect the dots
    grid_background : float
        The value to dilute the main color for background
    fontsize : int
        Set the fontsize for the plot
    add_intersections : bool, str, default: True
        Whether or which side to add the intersection size.
    add_sets_size : bool, str, default: True
        Whether or which side to add the sets size.
    add_labels : bool, str, default: True
        Whether or which side to add the label.
    width : float
    height : float
    size_scale : float
        The scale of the plot
    size_aspect : float
        The aspect ratio of the plot

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> from marsilea.upset import UpsetData, Upset
        >>> data = UpsetData.from_sets([[1, 2, 3, 4],
        >>>                             [3, 4, 5, 6],
        >>>                             [1, 6, 10, 11]])
        >>> Upset(data).render()


    """

    def __init__(
        self,
        data: UpsetData,
        orient="h",
        sort_sets=None,  # ascending, descending
        sets_order=None,
        sets_color=None,
        sort_subsets="cardinality",  # cardinality, degree
        min_degree=None,
        max_degree=None,
        min_cardinality=None,
        max_cardinality=None,
        color=".1",
        shading=0.3,
        radius=50,
        linewidth=1.5,
        grid_background=0.1,
        fontsize=None,
        add_intersections=True,
        add_sets_size=True,
        add_labels=True,
        width=None,
        height=None,
        size_scale=0.3,
        size_aspect=1.0,
    ):
        # The modification happens inplace
        upset_data = data
        upset_data.filter(
            min_degree=min_degree,
            max_degree=max_degree,
            min_cardinality=min_cardinality,
            max_cardinality=max_cardinality,
        )

        ascending = sort_subsets.startswith("-")
        if ascending:
            sort_subsets = sort_subsets[1:]
        upset_data.sort_subsets(by=sort_subsets, ascending=ascending)

        ascending = sort_sets == "ascending"
        upset_data.sort_sets(order=sets_order, ascending=ascending)

        sets_size = upset_data.sets_size()
        sets_table = upset_data.sets_table()

        self.sets_size = sets_size
        self.sets_table = sets_table

        self.data = upset_data
        self.color = color
        self.shading = shading
        self.linewidth = linewidth
        self.grid_background = grid_background
        self.fontsize = fontsize
        self.height = height
        self.width = width
        self.radius = radius
        if sets_color is None:
            sets_color = [self.color for _ in range(len(sets_size))]
        self.sets_color = np.asarray(sets_color)

        self._subset_styles = {}
        self._subset_line_styles = {}
        self._legend_entries = []
        self._add_intersections = add_intersections
        self._intersection_bar = None
        self._sets_size_bar = None

        if orient not in ["h", "v"]:
            raise ValueError("orient must be 'h' or 'v'")

        main_shape = sets_table.index.to_frame().shape
        if orient == "h":
            main_shape = main_shape[::-1]
        self.orient = orient

        width, height = get_canvas_size_by_data(
            main_shape,
            scale=size_scale,
            width=width,
            height=height,
            aspect=size_aspect,
        )

        super().__init__(width=width, height=height)
        if add_intersections:
            if isinstance(add_intersections, str):
                side = add_intersections
            else:
                side = "top" if orient == "h" else "right"
            self.add_intersections(side)
        if add_labels:
            if isinstance(add_labels, str):
                side = add_labels
            else:
                side = "right" if orient == "h" else "bottom"
            self.add_sets_label(side)
        if add_sets_size:
            if isinstance(add_sets_size, str):
                side = add_sets_size
            else:
                side = "left" if orient == "h" else "top"
            self.add_sets_size(side, color=self.sets_color)

    def highlight_subsets(
        self,
        present=None,
        absent=None,
        min_cardinality=None,
        max_cardinality=None,
        min_degree=None,
        max_degree=None,
        facecolor=None,
        edgecolor=None,
        edgewidth=None,
        hatch=None,
        edgestyle=None,
        label=None,
    ):
        """Highlight a subset of the data.

        Notice that the color of hatch is determined by the edgecolor.

        Parameters
        ----------
        present :
            The sets that present in the subset
        absent :
            The sets that absent in the subset
        min_cardinality :
            The minimum cardinality of the subset
        max_cardinality :
             The maximum cardinality of the subset
        min_degree :
            The minimum degree of the subset
        max_degree :
            The maximum degree of the subset
        facecolor :
            The facecolor to decorate the subset
        edgecolor :
            The color of the edge line
        edgewidth :
            The edgewidth of the edge line
        hatch :
            The fill pattern
        edgestyle :
            The style of edge line
        label :
            The label for the highlighting

        """
        marks = self.data.mark(
            present=present,
            absent=absent,
            min_cardinality=min_cardinality,
            max_cardinality=max_cardinality,
            min_degree=min_degree,
            max_degree=max_degree,
        )

        def _label(name, value):
            if value is not None:
                return f"{name}={value}"
            else:
                return ""

        if label is None:
            present = _label("present", present)
            absent = _label("absent", absent)
            min_cardinality = _label("min_cardinality", min_cardinality)
            max_cardinality = _label("max_cardinality", max_cardinality)
            min_degree = _label("min_degree", min_degree)
            max_degree = _label("max_degree", max_degree)

            label = ", ".join(
                [
                    s
                    for s in [
                        present,
                        absent,
                        min_cardinality,
                        max_cardinality,
                        min_degree,
                        max_degree,
                    ]
                    if s
                ]
            )

        styles = dict(
            facecolor=facecolor,
            edgecolor=edgecolor,
            linestyle=edgestyle,
            linewidth=edgewidth,
            hatch=hatch,
            label=label,
        )
        styles = {k: v for k, v in styles.items() if v is not None}

        line_styles = dict()
        linecolor = facecolor if facecolor is not None else None
        linecolor = edgecolor if edgecolor is not None else linecolor
        line_styles.update(color=linecolor)

        for i, m in enumerate(marks):
            if m:
                current_styles = styles.copy()
                sub_set_style = self._subset_styles.get(i)
                if sub_set_style is not None:
                    sub_set_style.update(current_styles)
                    self._subset_line_styles[i].update(line_styles)
                else:
                    self._subset_styles[i] = current_styles
                    self._subset_line_styles[i] = {**line_styles}

        if "facecolor" not in styles.keys():
            styles["facecolor"] = "none"
        if "edgecolor" not in styles.keys():
            styles["edgecolor"] = "none"
        self._legend_entries.append(styles)

    def _check_side(self, side, chart_name, allow):
        options = allow[self.orient]
        if side not in options:
            msg = (
                f"{chart_name} cannot be placed at '{side}', "
                f"try {' ,'.join(options)}"
            )
            raise ValueError(msg)

    def add_intersections(self, side, pad=0.1, size=1.0):
        self._add_intersections = True
        self._check_side(
            side, "Intersections", dict(h=["top", "bottom"], v=["left", "right"])
        )
        data = self.data.cardinality()
        self._intersection_bar = Numbers(data, color=self.color)
        self.add_plot(side, self._intersection_bar, size=size, pad=pad)

    def add_sets_size(self, side, pad=0.1, size=1.0, **props):
        self._check_side(
            side, "Sets size", dict(h=["left", "right"], v=["top", "bottom"])
        )
        data = self.sets_size
        options = dict(color=self.color)
        options.update(props)
        self._sets_size_bar = Numbers(data, **options)
        self.add_plot(side, self._sets_size_bar, size=size, pad=pad)

    def add_sets_label(self, side, pad=0.1, size=None, **props):
        self._check_side(
            side, "Sets label", dict(h=["left", "right"], v=["top", "bottom"])
        )
        data = self.data.sets_names
        self.add_plot(side, Labels(data, **props), pad=pad, size=size)

    def get_intersection_ax(self):
        return self.get_ax("Intersections")

    def get_sets_size_ax(self):
        return self.get_ax("Sets size")

    def get_sets_label_ax(self):
        return self.get_ax("Sets label")

    def get_data(self):
        return self.data

    _attr_plotter = {
        "bar": Bar,
        "box": Box,
        "boxen": Boxen,
        "violin": Violin,
        "point": Point,
        "strip": Strip,
        "swarm": Swarm,
        "stack_bar": StackBar,
        "number": Numbers,
    }

    @classmethod
    def update_attr_plotter(cls, attr_plotter):
        """Update the global upset plot for attr plotter"""
        cls._attr_plotter.update(attr_plotter)

    def add_sets_attr(
        self, side, attr_name, plot, name=None, pad=0.1, size=None, plot_kws=None
    ):
        """Add a plot for the sets attribute

        Parameters
        ----------
        side : str
            The side to add the plot, can be 'left', 'right', 'top', 'bottom'
        attr_name : str
            The name of the attribute
        plot : str
            The type of plot, can be 'bar', 'box', 'boxen', 'violin', 'point',
            'strip', 'swarm', 'stack_bar', 'number'
        name : str, optional
            The name of the plot
        pad : float, optional
            The padding between the plot and the axis
        size : float, optional
            The size of the plot
        plot_kws : dict, optional
            The keyword arguments for the plot

        """
        data = self.data.sets_attrs
        attr = data[attr_name]
        plot = self._attr_plotter[plot]
        kws = {"label": attr_name}
        if plot_kws is not None:
            kws.update(plot_kws)
        self.add_plot(side, plot(attr, **plot_kws), name=name, pad=pad, size=size)

    def add_items_attr(
        self, side, attr_name, plot, name=None, pad=0.1, size=None, plot_kws=None
    ):
        """Add a plot for the items attribute

        Parameters
        ----------
        side : str
            The side to add the plot, can be 'left', 'right', 'top', 'bottom'
        attr_name : str
            The name of the attribute
        plot : str
            The type of plot, can be 'bar', 'box', 'boxen', 'violin', 'point',
            'strip', 'swarm', 'stack_bar', 'number'
        name : str, optional
            The name of the plot
        pad : float, optional
            The padding between the plot and the axis
        size : float, optional
            The size of the plot
        plot_kws : dict, optional
            The keyword arguments for the plot

        """

        data_collector = self.data.get_items_attr(attr_name)

        construct = pd.DataFrame(data_collector).T

        if plot == StackBar:
            collect = [Counter(col) for col in data_collector]
            construct = pd.DataFrame(collect).T
            construct = (
                (construct.loc[~pd.isnull(construct.index)]).fillna(0).astype(int)
            )

        plot = self._attr_plotter[plot]
        kws = {"label": attr_name}
        if plot_kws is not None:
            kws.update(plot_kws)
        self.add_plot(side, plot(construct, **kws), name=name, pad=pad, size=size)

    def _render_matrix(self, ax):
        ax.set_axis_off()

        bg_circles = []
        lines, circles = [], []
        matrix = self.sets_table.index.to_frame().to_numpy()
        X, Y = matrix.shape
        xticks = np.arange(X)
        yticks = np.arange(Y)
        xv, yv = np.meshgrid(xticks, yticks)

        # draw bg
        if self.shading > 0:
            if self.orient == "v":
                xv, yv = yv, xv
            ax.scatter(
                xv,
                yv,
                s=self.radius,
                facecolor=self.color,
                alpha=self.shading,
                edgecolor="none",
            )

        for ix1, chunk in enumerate(matrix):
            custom_style = self._subset_styles.get(ix1)
            custom_line_style = self._subset_line_styles.get(ix1)
            if custom_style is None:
                custom_style = {}
                custom_line_style = {}

            cy = np.nonzero(chunk)[0]
            cx = np.repeat(ix1, len(cy))
            if len(cy) > 0:
                line_low, line_up = np.min(cy), np.max(cy)
                if (self.linewidth > 0) & (line_up - line_low > 0):
                    line_style = {
                        "color": self.color,
                        "lw": self.linewidth,
                        **custom_line_style,
                    }
                    xs, ys = ix1, (line_low, line_up)
                    liner = ax.vlines if self.orient == "h" else ax.hlines
                    liner(xs, *ys, **line_style)
                scatter_colors = self.sets_color[cy]
                if self.orient == "v":
                    cx, cy = cy, cx
                current_style = {
                    "facecolor": scatter_colors,
                    "zorder": 100,
                    "alpha": 1,
                    **custom_style,
                }
                ax.scatter(cx, cy, s=self.radius, **current_style)

        xlow, xup = 0 - 0.5, np.max(xv) + 0.5
        ylow, yup = np.max(yv) + 0.5, 0 - 0.5
        ax.set_xlim(xlow, xup)
        ax.set_ylim(ylow, yup)

        if self.orient == "h":
            bg_coords = zip(cycle([xlow]), np.arange(yup, ylow))
            height = 1
            width = xup - xlow
        else:
            bg_coords = zip(np.arange(xlow, xup), cycle([yup]))
            width = 1
            height = ylow - yup
        for i, coord in enumerate(bg_coords):
            if i % 2 == 0:
                rect = Rectangle(
                    xy=coord,
                    height=height,
                    width=width,
                    facecolor=self.color,
                    alpha=self.grid_background,
                )
                bg_circles.append(rect)
        # add bg_circles
        bg_circles = PatchCollection(bg_circles, match_original=True)
        ax.add_collection(bg_circles)

        # add lines
        for line in lines:
            ax.add_artist(line)

        # add circles
        circles = PatchCollection(circles, match_original=True)
        ax.add_collection(circles)

    def _extra_legends(self):
        handles = [Patch(**entry) for entry in self._legend_entries]
        highlight_legend = ListLegend(handles=handles, handlelength=2)
        # highlight_legend.set_figure(None)
        return {"highlight_subsets": [highlight_legend]}

    def render(self, figure=None, scale=1):
        super().render(figure=figure, scale=scale)
        main_ax = self.get_main_ax()
        self._render_matrix(main_ax)
        # apply highlight style to bar
        if self._add_intersections:
            for ix, rect in enumerate(self._intersection_bar.bars):
                bar_style = self._subset_styles.get(ix)
                if bar_style is not None:
                    rect.set(**bar_style)



import numpy as np
from legendkit import ListLegend
from matplotlib import pyplot as plt
from matplotlib.artist import Artist
from matplotlib.collections import PathCollection
from matplotlib.lines import Line2D
from matplotlib.markers import MarkerStyle
from matplotlib.patches import Rectangle, Polygon
from matplotlib.transforms import IdentityTransform
from typing import Mapping, Iterable

from .base import ClusterBoard
from .layout import close_ticks
from .plotter import RenderPlan


class LayersMesh(RenderPlan):
    """The mesh that draw customized elements in multi-layers

    LayersMesh is a powerful solution for drawing and visualizing customized
    elements in multi-layered formats.
    It offers two flexible ways to specify layers:

    #. A single layer of data containing different elements.
    #. Multiple layers of data, where each layer represents a customized element, \
    overlaying elements on top of one another.

    When supplying multiple layers,
    the drawing order follows the order of the input array.
    Override this order by controlling the `zorder` attribute in your :class:`Pieces`.

    Parameters
    ----------

    data : np.ndarray, optional
        Use this parameter for a single layer
    layers : list of data
        For multiple layers, use this parameter.
        Each layer must be a boolean matrix indicating
        whether the element is rendered at a specific cell.
    pieces : dict or array
        For a single layer, use a dict to define how to render each element.
        For multiple layers, use an array to define the rendering of each layer.
    label : str
        The mesh label, displayed only when added to the side plot.
    label_loc : str
        The label location.
    props : dict
        See :class:`matplotlib.text.Text`


    Examples
    --------

    Drawing a single layer:

    .. plot::
        :context: close-figs

        >>> from marsilea.layers import LayersMesh, Rect, FrameRect
        >>> data = np.random.choice([1, 2, 3], (10, 10))
        >>> pieces = {1: Rect(color="r", label="1"),
        ...           2: FrameRect(color="g", label="2"),
        ...           3: Rect(color="b", label="3")}
        >>> _, ax = plt.subplots()
        >>> LayersMesh(data=data, pieces=pieces).render(ax)

    Drawing multiple layers:

    .. plot::
        :context: close-figs

        >>> d1 = np.random.choice([1, 2, 3], (10, 10))
        >>> d2 = np.random.choice([1, 2, 3], (10, 10))
        >>> d3 = np.random.choice([1, 2, 3], (10, 10))
        >>> pieces = [Rect(color="r", label="1"),
        ...           FrameRect(color="g", label="2"),
        ...           Rect(color="b", label="3")]
        >>> _, ax = plt.subplots()
        >>> LayersMesh(layers=[d1, d2, d3], pieces=pieces).render(ax)


    """

    render_main = True

    def __init__(
        self,
        data=None,
        layers=None,
        pieces=None,
        shrink=(1.0, 1.0),
        label=None,
        label_loc=None,
        props=None,
        legend_kws=None,
    ):
        # render one layer
        # with different elements
        if data is not None:
            if not isinstance(pieces, Mapping):
                msg = f"Expect pieces to be dict " f"but found {type(pieces)} instead."
                raise TypeError(msg)
            self.pieces_mapper = pieces
            self.mode = "cell"
            self.set_data(data)
            self._one_layer = True
        # render multiple layers
        # each layer is an elements
        else:
            if not isinstance(pieces, Iterable):
                msg = f"Expect pieces to be list " f"but found {type(pieces)} instead."
                raise TypeError(msg)
            self.pieces, data = self._sort_by_zorder(pieces, layers)
            self.mode = "layer"
            self.set_data(*data)
            self._one_layer = len(data) == 1
        self.x_offset = (1 - shrink[0]) / 2
        self.y_offset = (1 - shrink[1]) / 2
        self.width = shrink[0]
        self.height = shrink[1]
        self.label = label
        self.label_loc = label_loc
        self.props = props
        default_legend_kws = dict(frameon=False, handlelength=1, handleheight=1)
        if legend_kws is not None:
            default_legend_kws.update(legend_kws)
        self._legend_kws = default_legend_kws

    @staticmethod
    def _sort_by_zorder(pieces, layers):
        ix_pieces = sorted(enumerate(pieces), key=lambda x: x[1].zorder)
        sorted_pieces = []
        sorted_layers = []
        for ix, piece in ix_pieces:
            sorted_pieces.append(piece)
            sorted_layers.append(layers[ix])
        return sorted_pieces, sorted_layers

    def get_legends(self):
        if self.mode == "cell":
            handles = list(self.pieces_mapper.values())
        else:
            handles = self.pieces
        new_handles = []
        labels = []
        handler_map = {}
        for h in handles:
            if h.legend_entry:
                new_handles.append(h)
                labels.append(h.get_label())
                handler_map[h] = h
        return ListLegend(
            handles=new_handles,
            labels=labels,
            handler_map=handler_map,
            **self._legend_kws,
        )

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data

        if self.mode == "layer":
            if self.is_flank:
                data = [d.T for d in data]
            if self._one_layer:
                data = [data]
            Y, X = data[0].shape
        else:
            if self.is_flank:
                data = data.T
            Y, X = data.shape
        ax.set_axis_off()
        ax.set_xlim(0, X)
        ax.set_ylim(0, Y)
        if self.mode == "layer":
            for layer, piece in zip(data, self.pieces):
                for iy, row in enumerate(layer):
                    for ix, v in enumerate(row):
                        if v:
                            art = piece.draw(
                                ix + self.x_offset,
                                iy + self.y_offset,
                                self.width,
                                self.height,
                                ax,
                            )
                            ax.add_artist(art)
        else:
            for iy, row in enumerate(data):
                for ix, v in enumerate(row):
                    piece = self.pieces_mapper[v]
                    art = piece.draw(
                        ix + self.x_offset,
                        iy + self.y_offset,
                        self.width,
                        self.height,
                        ax,
                    )
                    ax.add_artist(art)
        ax.invert_yaxis()


class Layers(ClusterBoard):
    def __init__(
        self,
        data=None,
        layers=None,
        pieces=None,
        cluster_data=None,
        shrink=(1.0, 1.0),
        height=None,
        width=None,
        aspect=1.0,
        name=None,
    ):
        mesh = LayersMesh(data=data, layers=layers, pieces=pieces, shrink=shrink)
        if cluster_data is None:
            self._allow_cluster = False
            data_shape = mesh.get_data()[0].shape
            # create numeric data explicitly
            # in case user input string data
            cluster_data = np.random.randn(*data_shape)
        # if (width is not None) & (height is not None):
        #     main_aspect = height / width
        # else:
        #     Y, X = cluster_data.shape
        #     main_aspect = Y * aspect / X
        super().__init__(cluster_data, width=width, height=height, name=name)

        self.add_layer(mesh)


class Piece:
    label = None
    legend_entry = True
    zorder = 0
    color = "C0"

    def get_label(self):
        if self.label is None:
            return self.__class__.__name__
        else:
            return self.label

    def set_label(self, label):
        self.label = label

    @staticmethod
    def draw_center(x, y, w, h):
        cx = x + w / 2.0
        cy = y + h / 2.0
        return cx, cy

    def draw(self, x, y, w, h, ax) -> Artist:
        raise NotImplementedError

    def legend(self, x, y, w, h) -> Artist:
        return self.draw(x, y, w, h, None)

    def legend_artist(self, legend, orig_handle, fontsize, handlebox):
        x, y = handlebox.xdescent, handlebox.ydescent
        w, h = handlebox.width, handlebox.height
        p = self.legend(-x, -y, w, h)
        handlebox.add_artist(p)
        return p


class Rect(Piece):
    def __init__(self, color="C0", label=None, legend=True, zorder=0):
        self.color = color
        self.label = label
        self.legend_entry = legend
        self.zorder = zorder

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(color='{self.color}', " f"label='{self.label}')"
        )

    def draw(self, x, y, w, h, ax) -> Artist:
        return Rectangle((x, y), w, h, facecolor=self.color)


class FracRect(Piece):
    def __init__(self, color="C0", frac=(0.9, 0.5), label=None, legend=True, zorder=0):
        self.color = color
        self.label = label
        self.frac = frac
        self.legend_entry = legend
        self.zorder = zorder

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(color='{self.color}', " f"label='{self.label}')"
        )

    def draw(self, x, y, w, h, ax):
        fx, fy = self.frac
        draw_w, draw_h = w * fx, h * fy
        # compute the actual drawing bbox
        # Lower-left corner
        draw_x = x + (w - draw_w) / 2.0
        draw_y = y + (h - draw_h) / 2.0
        return Rectangle((draw_x, draw_y), draw_w, draw_h, fc=self.color)


class FrameRect(Piece):
    def __init__(self, color="C0", width=1, label=None, legend=True, zorder=0):
        self.color = color
        self.width = width
        self.label = label
        self.legend_entry = legend
        self.zorder = zorder

    def __repr__(self):
        return f"{self.__class__.__name__}(color={self.color}, " f"label={self.label})"

    def draw(self, x, y, w, h, ax):
        return Rectangle((x, y), w, h, fill=False, ec=self.color, linewidth=self.width)


class RightTri(Piece):
    point_order = {
        "lower left": [1, 0, 3],
        "lower right": [2, 0, 3],
        "upper left": [0, 1, 2],
        "upper right": [1, 2, 3],
    }

    def __init__(
        self, color="C0", right_angle="lower left", label=None, legend=True, zorder=0
    ):
        self.color = color
        self.pos = right_angle
        self.label = label
        self.legend_entry = legend
        self.zorder = zorder

    def __repr__(self):
        return f"{self.__class__.__name__}(color={self.color}, " f"label={self.label})"

    def draw(self, x, y, w, h, ax):
        p0 = (x, y)
        p1 = (x, y + h)
        p2 = (x + w, y + h)
        p3 = (x + w, y)
        points = np.array([p0, p1, p2, p3])
        ps = points[self.point_order[self.pos]]
        return Polygon(ps, fc=self.color)


class Marker(Piece):
    def __init__(self, marker, color="C0", size=32, label=None, legend=True, zorder=0):
        self.color = color
        self.label = label
        self.legend_entry = legend
        self.zorder = zorder
        self.size = size
        m = MarkerStyle(marker)
        self.path = m.get_path().transformed(m.get_transform())
        self.marker = marker

    def draw(self, x, y, w, h, ax):
        c = self.draw_center(x, y, w, h)
        collection = PathCollection(
            (self.path,),
            [self.size],
            offsets=[c],
            offset_transform=ax.transData,
            facecolors=self.color,
        )
        collection.set_transform(IdentityTransform())

        return collection

    def legend(self, x, y, w, h):
        return Line2D(
            [0], [0], marker=self.marker, color=self.color, markersize=self.size
        )


def preview(piece: Piece, figsize=(1, 1)):
    figure = plt.figure(figsize=figsize)
    ax = figure.add_axes([0, 0, 1, 1])
    close_ticks(ax)
    arts = piece.draw(0, 0, 1, 1, ax)
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.add_artist(arts)



from __future__ import annotations

import logging
import numpy as np
import pandas as pd

from .base import ClusterBoard
from .plotter import ColorMesh, SizedMesh, Colors
from .utils import get_plot_name

log = logging.getLogger("marsilea")


class Heatmap(ClusterBoard):
    """Heatmap

    See :class:`ColorMesh <marsilea.plotter.ColorMesh>` for details

    Other Parameters
    ----------------
    cluster_data : matrix data
        You can override the data used to perform cluster
        By default will use the plotting data.
    name : str
        The name of this Heatmap
    width, height : float
        The size in inches to define the size of main canvas

    """

    def __init__(
        self,
        data: np.ndarray,
        vmin=None,
        vmax=None,
        cmap=None,
        norm=None,
        center=None,
        mask=None,
        alpha=None,
        linewidth=0,
        linecolor="white",
        annot=None,
        fmt=None,
        annot_kws=None,
        label=None,
        cbar_kws=None,
        name=None,
        width=None,
        height=None,
        cluster_data=None,
        init_main=True,
        legend=True,
        **kwargs,
    ):
        if cluster_data is None:
            if isinstance(data, pd.DataFrame):
                cluster_data = data.values
            else:
                cluster_data = data
        super().__init__(
            cluster_data, width=width, height=height, name=name, init_main=init_main
        )
        mesh = ColorMesh(
            data,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            norm=norm,
            center=center,
            mask=mask,
            alpha=alpha,
            linewidth=linewidth,
            linecolor=linecolor,
            annot=annot,
            fmt=fmt,
            annot_kws=annot_kws,
            label=label,
            cbar_kws=cbar_kws,
            **kwargs,
        )
        name = get_plot_name(name, "main", mesh.__class__.__name__)
        mesh.set(name=name)
        self.add_layer(mesh, legend=legend)


class CatHeatmap(ClusterBoard):
    """Categorical Heatmap

    See :class:`Colors <marsilea.plotter.Colors>` for details

    Other Parameters
    ----------------
    name : str
        The name of this Heatmap
    cluster_data : matrix data
        You can override the data used to perform cluster
        By default will use the plotting data.
    width, height : float
        The size in inches to define the size of main canvas

    """

    def __init__(
        self,
        data,
        palette=None,
        cmap=None,
        mask=None,
        name=None,
        width=None,
        height=None,
        legend=True,
        cluster_data=None,
        linewidth=0,
        linecolor="white",
        **kwargs,
    ):
        mesh = Colors(
            data,
            palette=palette,
            cmap=cmap,
            mask=mask,
            linewidth=linewidth,
            linecolor=linecolor,
            **kwargs,
        )
        if cluster_data is None:
            cluster_data = mesh.cluster_data
        super().__init__(cluster_data, width=width, height=height, name=name)
        name = get_plot_name(name, "main", mesh.__class__.__name__)
        mesh.set(name=name)
        self.add_layer(mesh, legend=legend)


class SizedHeatmap(ClusterBoard):
    """Sized Heatmap

    See :class:`SizedMesh <marsilea.plotter.SizedMesh>` for details

    Other Parameters
    ----------------
    name : str
        The name of this Heatmap
    cluster_data : matrix data
        You can override the data used to perform cluster
        By default will use the size data.
    width, height : float
        The size in inches to define the size of main canvas

    """

    def __init__(
        self,
        size,
        color=None,
        cluster_data=None,
        name=None,
        width=None,
        height=None,
        legend=True,
        **kwargs,
    ):
        if cluster_data is None:
            if isinstance(size, pd.DataFrame):
                cluster_data = size.values
            else:
                cluster_data = size
        super().__init__(cluster_data, width=width, height=height, name=name)

        mesh = SizedMesh(size=size, color=color, **kwargs)
        self.add_layer(mesh, legend=legend)



import pandas as pd
import numpy as np
from platformdirs import user_cache_path
from urllib.request import urlretrieve

NAME = "Marsilea"
BASE_URL = "https://raw.githubusercontent.com/Marsilea-viz/marsilea-data/main"


def load_data(name, cache=True):
    """To load marsilea dataset

    - `cooking_oils`: Cooking oils dataset
    - `imdb`: The IMDB Top 100 Movies
    - `pbmc3k`: single-cell RNA-seq dataset from 10X Genomics
    - `oncoprint`: Subsample of Breast Invasive Carcinoma (TCGA, PanCancer Atlas)
    - `mouse_embryo`: Spatial mapping of mouse embryo at E12.5
    - `seq_align`: Sequence alignment data
    - `les_miserables`: Characters in Les Misérables
    - `sc_multiomics`: Single-cell multi-omics dataset from COVID-19 patients
    - `track`: Track data from GEO GSE137105 for MYC gene

    Parameters
    ----------
    name : str
        The name of the dataset
    cache : bool
        Download data and cache locally

    Returns
    -------

    """
    if name == "imdb":
        return _load_imdb(cache)
    elif name == "pbmc3k":
        return _load_pbmc3k(cache)
    elif name == "sc_multiomics":
        return _load_sc_multiomics(cache)
    elif name == "oncoprint":
        return _load_oncoprint(cache)
    elif name == "mouse_embryo":
        return _load_mouse_embryo(cache)
    elif name == "seq_align":
        return _load_sequence_alignment(cache)
    elif name == "cooking_oils":
        return _load_cooking_oils(cache)
    elif name == "les_miserables":
        return _load_les_miserables(cache)
    elif name == "track":
        return _load_track(cache)
    else:
        raise NameError("Dataset not found")


def _cache_remote(files, cache=True):
    if isinstance(files, str):
        files = [files]
    data_dir = user_cache_path(appname=NAME)
    data_dir.mkdir(exist_ok=True, parents=True)

    download_files = [f.replace("/", "_") for f in files]

    for fname, dfname in zip(files, download_files):
        url = f"{BASE_URL}/{fname}"
        # Will not download if cache and file exist
        dest = data_dir / dfname
        # Not download when cache and exist
        if not (cache and dest.exists()):
            try:
                urlretrieve(url, dest)
            except Exception as e:
                print(url)
                raise e

    if len(files) == 1:
        return data_dir / download_files[0]
    else:
        return [data_dir / f for f in download_files]


def _load_imdb(cache=True):
    imdb = _cache_remote("imdb.csv", cache=cache)
    return pd.read_csv(imdb)


def _load_pbmc3k(cache=True):
    exp, pct_cells, count = _cache_remote(
        ["pbmc3k/exp.csv", "pbmc3k/pct_cells.csv", "pbmc3k/count.csv"], cache=cache
    )
    return {
        "exp": pd.read_csv(exp, index_col=0),
        "pct_cells": pd.read_csv(pct_cells, index_col=0),
        "count": pd.read_csv(count, index_col=0),
    }


def _load_sc_multiomics(cache=True):
    stack, interaction = _cache_remote(
        [
            "sc-multiomics/sc-multiomics.npz",
            "sc-multiomics/sc-multiomics-interaction.csv",
        ]
    )

    dataset = np.load(stack, allow_pickle=True)
    interaction = pd.read_csv(interaction)

    data = {}
    for key in dataset.files:
        data[key] = dataset[key]
    data["interaction"] = interaction

    return data


def _load_oncoprint(cache=True):
    cna, mrna, methyl, clinical = _cache_remote(
        [
            "oncoprint/cna.csv",
            "oncoprint/mrna_exp.csv",
            "oncoprint/methyl_exp.csv",
            "oncoprint/clinical.csv",
        ],
        cache=cache,
    )
    return {
        "cna": pd.read_csv(cna, index_col=0),
        "mrna_exp": pd.read_csv(mrna, index_col=0),
        "methyl_exp": pd.read_csv(methyl, index_col=0),
        "clinical": pd.read_csv(clinical, index_col=0),
    }


def _load_mouse_embryo(cache=True):
    data = _cache_remote("mouse_embryo_E12.5.csv.gz", cache=cache)
    return pd.read_csv(data)


def _load_sequence_alignment(cache=True):
    data = _cache_remote("sequence_alignment.csv", cache=cache)
    return pd.read_csv(data, index_col=0)


def _load_cooking_oils(cache=True):
    data = _cache_remote("cooking_oils.csv", cache=cache)
    return pd.read_csv(data, index_col=0)


def _load_les_miserables(cache=True):
    nodes, links = _cache_remote(
        ["les-miserables/miserables_nodes.csv", "les-miserables/miserables_links.csv"],
        cache=cache,
    )
    return {"nodes": pd.read_csv(nodes), "links": pd.read_csv(links)}


def _load_track(cache=True):
    data = _cache_remote("track.parquet", cache=cache)
    return pd.read_parquet(data)



import numpy as np
import pandas as pd
from matplotlib.patches import PathPatch
from matplotlib.textpath import TextPath
from matplotlib.transforms import Affine2D, Bbox

from .base import StatsBase
from ..utils import pairwise, ECHARTS16


def path_char(
    pos,
    extend,
    t,
    ax,
    width=1.0,
    flip=False,
    mirror=False,
    direction="h",
    prop=None,
    usetex=False,
    **kwargs,
):
    w = width
    h = extend[1] - extend[0]

    x = pos - w / 2
    y = extend[0]

    if prop is None:
        prop = dict(family="monospace")
    tmp_path = TextPath((0, 0), t, size=1, prop=prop, usetex=usetex)

    if flip:
        transformation = Affine2D().scale(sx=1, sy=-1)
        tmp_path = transformation.transform_path(tmp_path)

    if mirror:
        transformation = Affine2D().scale(sx=-1, sy=1)
        tmp_path = transformation.transform_path(tmp_path)

    if direction == "v":
        w, h = h, w
        x, y = y, x

        rotate = Affine2D().rotate_deg(90)
        tmp_path = rotate.transform_path(tmp_path)

    bbox = Bbox.from_bounds(x, y, w, h)
    tmp_bbox = tmp_path.get_extents()

    hs = bbox.width / tmp_bbox.width
    vs = bbox.height / tmp_bbox.height

    if direction == "h":
        char_width = hs * tmp_bbox.width
        char_shift = (bbox.width - char_width) / 2.0
        tx = bbox.xmin + char_shift
        ty = bbox.ymin

    else:
        char_height = vs * tmp_bbox.height
        char_shift = (bbox.height - char_height) / 2.0
        tx = bbox.xmin
        ty = bbox.ymin + char_shift

    transformation = (
        Affine2D()
        .translate(tx=-tmp_bbox.xmin, ty=-tmp_bbox.ymin)
        .scale(sx=hs, sy=vs)
        .translate(tx=tx, ty=ty)
    )

    char_path = transformation.transform_path(tmp_path)

    patch = PathPatch(char_path, **kwargs)
    ax.add_patch(patch)


class SeqLogo(StatsBase):
    """Sequence logo

    Parameters
    ----------
    matrix : pandas.DataFrame
        Drawing matrix, the data is the height of logo
        with columns as positions and rows as nucleotides/amino acids letters.
    width : float, optional
        The width of each letter, by default .9, should be within [0, 1]
    color_encode : dict, optional
        The color encoding for each letter,
        should be a dict with key as letter and value as color
    stack : {"descending", "ascending", "normal"}, default: "descending"
        The stacking order of letters, by height. If normal will stack
        as the order in matrix.

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> import pandas as pd
        >>> from marsilea.plotter import SeqLogo
        >>> matrix = pd.DataFrame(
        ...     data=np.random.randint(1, 10, (4, 10)), index=list("ACGT")
        ... )
        >>> _, ax = plt.subplots()
        >>> colors = {"A": "r", "C": "b", "G": "g", "T": "black"}
        >>> SeqLogo(matrix, color_encode=colors).render(ax)

    """

    render_main = False

    def __init__(
        self,
        matrix: pd.DataFrame,
        width=0.9,
        color_encode=None,
        stack="descending",  # "descending", "ascending", "normal"
        **kwargs,
    ):
        self.matrix = matrix
        self.letters = matrix.index.to_numpy()
        self.set_data(matrix.to_numpy())
        if color_encode is None:
            color_encode = dict(zip(self.letters, ECHARTS16))
        self.color_encode = color_encode
        self.width = width
        self.stack = stack
        self.options = kwargs

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data

        direction = "h" if self.is_body else "v"
        flip = self.side != "top"
        mirror = self.side == "left"
        lim = np.max(data.sum(axis=0))
        for i, col in enumerate(data.T):
            letters = self.letters
            if self.stack != "normal":
                ix = np.argsort(col)
                if self.stack == "ascending":
                    ix = ix[::-1]
                letters = letters[ix]
                col = col[ix]

            extends = [0] + list(np.cumsum(col))
            pos = i + 0.5
            for t, extend in zip(letters, pairwise(extends)):
                facecolor = self.color_encode[t]
                options = {"facecolor": facecolor, "edgecolor": "none", **self.options}
                path_char(
                    pos,
                    extend,
                    t,
                    ax,
                    flip=flip,
                    mirror=mirror,
                    width=self.width,
                    direction=direction,
                    **options,
                )
        if self.is_body:
            ax.set_xlim(0, data.shape[1])
            ax.set_ylim(0, lim)
        else:
            ax.set_xlim(0, lim)
            ax.set_ylim(0, data.shape[1])
        if self.is_flank:
            ax.invert_yaxis()
        if self.side == "left":
            ax.invert_xaxis()
        if self.side == "bottom":
            ax.invert_yaxis()
        ax.set_axis_off()



import numpy as np

from marsilea.plotter.base import StatsBase


class Area(StatsBase):
    """Area plot

    Parameters
    ----------
    data : array-like
        The data to be plotted.
    color : color-like
        The color of the area
    add_outline : bool
        Whether to add outline to the area
    alpha : float
        The transparency of the area
    linecolor : color-like
        The color of the outline
    linewidth : float
        The width of the outline
    group_kws : dict
        The configurations that apply to each group
    **kwargs :
        Additional configurations for the area plot, \
        see :func:`matplotlib.pyplot.fill_between`

    Examples
    --------
    .. plot::
        :context: close-figs

        import numpy as np
        import matplotlib.pyplot as plt

        from marsilea.plotter import Area

        _, ax = plt.subplots()
        data = np.random.randint(0, 10, 10) + 1
        Area(data).render(ax)

    """

    def __init__(
        self,
        data,
        color=None,
        add_outline=True,
        alpha=0.4,
        linecolor=None,
        linewidth=1,
        group_kws=None,
        label=None,
        label_loc=None,
        label_props=None,
        **kwargs,
    ):
        if color is None:
            color = "skyblue"
        if linecolor is None:
            linecolor = "Slateblue"

        self.color = color
        self.add_outline = add_outline
        self.alpha = alpha
        self.linecolor = linecolor
        self.linewidth = linewidth
        self.kws = kwargs

        self.set_data(data)
        self.set_label(label, label_loc, label_props)
        if group_kws is not None:
            self.set_group_params(group_kws)

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data
        gp = spec.group_params
        if gp is None:
            gp = {}

        fill_options = {"color": self.color, "alpha": self.alpha, **self.kws, **gp}
        line_options = {"color": self.linecolor, "linewidth": self.linewidth, **gp}

        x = np.arange(len(data))
        if self.get_orient() == "h":
            ax.fill_betweenx(x, data, **fill_options)
            if self.add_outline:
                ax.plot(data, x, **line_options)
            ax.set_ylim(-0.5, len(data) - 0.5)
            if self.side == "left":
                ax.invert_xaxis()
        else:
            ax.fill_between(x, data, **fill_options)
            if self.add_outline:
                ax.plot(x, data, **line_options)
            ax.set_xlim(-0.5, len(data) - 0.5)
        return ax


# TODO: Implement StackArea
class StackArea(StatsBase):
    pass



import pandas as pd
import seaborn
from legendkit import CatLegend
from seaborn import color_palette
from typing import Mapping, Sequence

from .base import StatsBase
from ..utils import ECHARTS16


class _SeabornBase(StatsBase):
    _seaborn_plot = None
    datasets = None
    hue = None
    data = None

    def __init__(
        self,
        data,
        hue_order=None,
        palette=None,
        orient=None,
        legend_kws=None,
        group_kws=None,
        label=None,
        label_loc=None,
        label_props=None,
        **kwargs,
    ):
        if isinstance(data, Mapping):
            datasets = []
            self.hue = []
            if hue_order is None:
                hue_order = data.keys()
            for name in hue_order:
                self.hue.append(name)
                datasets.append(self.data_validator(data[name]))
            if isinstance(palette, Mapping):
                self.palette = palette
            else:
                if palette is None:
                    colors = ECHARTS16
                else:
                    colors = color_palette(palette, as_cmap=False)
                self.palette = dict(zip(self.hue, colors))
            kwargs["palette"] = self.palette
            self.set_data(*datasets)
        else:
            data = self.data_validator(data)
            self.set_data(data)
            # kwargs.setdefault('color', 'C0')
            # if (palette is None) and ('color' not in kwargs):
            #     kwargs['palette'] = "dark:C0"
            if palette is not None:
                kwargs["palette"] = palette
                if isinstance(palette, Sequence):
                    self.set_params({"palette": palette})

        kwargs.pop("x", None)
        kwargs.pop("y", None)
        kwargs.pop("hue", None)
        # kwargs.pop("orient", None)
        kwargs.pop("ax", None)
        self.kws = kwargs

        self.orient = orient
        self.set_label(label, label_loc, label_props)
        self.legend_kws = {} if legend_kws is None else legend_kws
        if group_kws is not None:
            self.set_group_params(group_kws)

    def get_legends(self):
        if self.hue is not None:
            labels = []
            colors = []
            for label, color in self.palette.items():
                labels.append(label)
                colors.append(color)
            options = dict(handle="square", size=1, draw=False)
            options.update(self.legend_kws)
            return CatLegend(colors=colors, labels=labels, **options)

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data
        gp = spec.group_params

        if gp is None:
            gp = {}
        x, y = "var", "value"
        if self.hue is not None:
            dfs = []
            for d, hue in zip(data, self.hue):
                df = pd.DataFrame(d)
                df = df.melt(var_name="var", value_name="value")
                df["hue"] = hue
                dfs.append(df)

            pdata = pd.concat(dfs).reset_index(drop=True)
            self.kws["hue"] = "hue"
            self.kws["hue_order"] = self.hue
            if self.get_orient() == "h":
                x, y = y, x
            self.kws["x"] = x
            self.kws["y"] = y
            options = {**self.kws, **gp}

        else:
            pdata = pd.DataFrame(data).melt(var_name="var", value_name="value")
            if self.get_orient() == "h":
                x, y = y, x
            self.kws["x"] = x
            self.kws["y"] = y
            if spec.params is not None:
                palette = [p.get("palette", "C0") for p in spec.params]
                self.kws["palette"] = palette
            options = {**self.kws, **gp}
            if options.get("palette") is not None:
                options["hue"] = "var"

        orient = self.get_orient()
        if self.side == "left":
            if not ax.xaxis_inverted():
                ax.invert_xaxis()
        # barplot(data=data, orient=orient, ax=ax, **self.kws)
        plotter = getattr(seaborn, self._seaborn_plot)
        plotter(data=pdata, orient=orient, ax=ax, **options)
        ax.set(xlabel=None, ylabel=None)
        leg = ax.get_legend()
        if leg is not None:
            leg.remove()


def _seaborn_doc(obj: _SeabornBase):
    cls_name = obj.__name__

    sdata = "np.random.rand(10, 10)"
    hue_data = "{'a': sdata, 'b': sdata}"
    kws = "color='#DB4D6D'"
    h_kws = "group_kws={'color': colors}"

    if cls_name == "Swarm":
        sdata = "np.random.rand(50, 10)"
        kws = "color='#DB4D6D', size=2, dodge=True"
        h_kws = "group_kws={'color': colors}, size=2"

    elif cls_name == "Strip":
        sdata = "np.random.rand(50, 10)"
        kws = "color='#DB4D6D', size=2, dodge=True"
        h_kws = "group_kws={'color': colors}, size=2"

    elif cls_name == "Point":
        hue_data = "{'a': sdata, 'b': sdata * 2}"

    base_doc = f"""Wrapper for seaborn's {obj._seaborn_plot}
    
    .. note::
        .. rubric:: About data format
        
        You can only use wide-format for this plot, the number of columns
        of your input data should match your main data, this allow the data
        to be split and reorder if split and cluster is applied.
        
    
    Parameters
    ----------
    data : np.ndarray, pd.DataFrame
        The wide-format data. To input 'hue' like data, 
        you need to input a dict.
        eg: :code:`{{'hue1': data1, 'hue2': data2}}`.
    hue_order : array of str
        The order of hue
    palette : dict of label, color
    label : str
        The label of your data
    legend_kws : dict
        Configurations for legend
    group_kws : dict
        Configurations that apply to each group, should be something like
        :code:`{{'colors': ['C0', 'C1', 'C2']}}` if you have three groups.
    kwargs : 
        See :func:`seaborn.{obj._seaborn_plot}`
        
    Examples
    --------
    
    To render seaborn plots as side plots
    
    .. plot::
        :context: close-figs
        
        >>> import marsilea as ma
        >>> from marsilea.plotter import {cls_name}
        >>> data = np.random.randn(10, 10)
        >>> sdata = {sdata}
        >>> plot = {cls_name}(sdata, {kws})
        >>> h = ma.Heatmap(data)
        >>> h.cut_rows(cut=[3, 7])
        >>> h.add_right(plot)
        >>> h.render()
    """

    extend_examples = f"""
    It's possible to add hue data
    
    .. plot::
        :context: close-figs
        
        >>> plot = {cls_name}({hue_data}, {kws})
        >>> h = ma.Heatmap(data)
        >>> h.cut_rows(cut=[3, 7])
        >>> h.add_right(plot)
        >>> h.render()
        
    You can also draw it on the main canvas
    
    .. plot::
        :context: close-figs
        
        >>> plot = {cls_name}(sdata, {kws})
        >>> colors = ['#66327C', '#FFB11B', '#A8D8B9']
        >>> anno = ma.plotter.Chunk(['C1', 'C2', 'C3'], colors, padding=10)
        >>> cb = ma.ClusterBoard(data, height=2, margin=.5)
        >>> cb.add_layer(plot)
        >>> cb.cut_cols([3, 7])
        >>> cb.add_bottom(anno)
        >>> cb.render()
        
    To layout in a different orient and style each group
    
    .. plot::
        :context: close-figs
        
        >>> plot = {cls_name}(sdata, orient='h',
        ...                   {h_kws})
        >>> anno = ma.plotter.Chunk(['C1', 'C2', 'C3'], colors, padding=10)
        >>> cb = ma.ClusterBoard(data.T, width=2)
        >>> cb.add_layer(plot)
        >>> cb.cut_rows([3, 7])
        >>> cb.add_left(anno)
        >>> cb.render()
        
    """
    if cls_name == "Count":
        obj.__doc__ = base_doc
    else:
        obj.__doc__ = base_doc + extend_examples
    return obj


@_seaborn_doc
class Bar(_SeabornBase):
    _seaborn_plot = "barplot"


@_seaborn_doc
class Box(_SeabornBase):
    _seaborn_plot = "boxplot"


@_seaborn_doc
class Boxen(_SeabornBase):
    _seaborn_plot = "boxenplot"


@_seaborn_doc
class Violin(_SeabornBase):
    _seaborn_plot = "violinplot"


@_seaborn_doc
class Point(_SeabornBase):
    _seaborn_plot = "pointplot"


# @_seaborn_doc
# class Count(_SeabornBase):
#     _seaborn_plot = "countplot"


@_seaborn_doc
class Strip(_SeabornBase):
    _seaborn_plot = "stripplot"


@_seaborn_doc
class Swarm(_SeabornBase):
    _seaborn_plot = "swarmplot"



"""Declarative creation of composable visualization"""

__version__ = "0.5.0"

import marsilea.plotter as plotter
from ._deform import Deformation
from .base import (
    WhiteBoard,
    ClusterBoard,
    ZeroWidth,
    ZeroHeight,
    ZeroWidthCluster,
    ZeroHeightCluster,
    CompositeBoard,
    StackBoard,
)
from .dataset import load_data
from .dendrogram import Dendrogram, GroupDendrogram
from .heatmap import Heatmap, SizedHeatmap, CatHeatmap
from .layers import Piece, Layers
from .layout import CrossLayout, CompositeCrossLayout, StackCrossLayout
from .upset import UpsetData, Upset



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import List

import marsilea.layers as mlayers
import numpy as np
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle

# The preset follows the style of cBioPortal as much as possible
# https://github.com/cBioPortal/cbioportal-frontend/blob/master/src/shared/components/oncoprint/geneticrules.ts
MUT_COLOR_MISSENSE = "#008000"
MUT_COLOR_MISSENSE_PASSENGER = "#53D400"
MUT_COLOR_INFRAME = "#993404"
MUT_COLOR_INFRAME_PASSENGER = "#a68028"
MUT_COLOR_TRUNC = "#000000"
MUT_COLOR_TRUNC_PASSENGER = "#708090"
MUT_COLOR_SPLICE = "#e5802b"
MUT_COLOR_SPLICE_PASSENGER = "#f0b87b"
MUT_COLOR_PROMOTER = "#00B7CE"
MUT_COLOR_OTHER = "#cf58bc"

MUT_DRIVER = "#000000"
MUT_VUS = "#696969"
MUT_COLOR_GERMLINE = "#FFFFFF"

MRNA_COLOR_HIGH = "#ff9999"
MRNA_COLOR_LOW = "#6699cc"

PROT_COLOR_HIGH = "#ff3df8"
PROT_COLOR_LOW = "#00E1FF"

CNA_COLOR_AMP = "#ff0000"
CNA_COLOR_GAIN = "#ffb6c1"
CNA_COLOR_HETLOSS = "#8fd8d8"
CNA_COLOR_HOMDEL = "#0000ff"

MUT_COLOR_FUSION = "#C900A1"
STRUCTURAL_VARIANT_COLOR = "#8B00C9"
STRUCTURAL_VARIANT_PASSENGER_COLOR = "#ce92e8"

DEFAULT_GREY = "#BEBEBE"


class Alteration(Enum):
    BACKGROUND = auto()
    AMP = auto()
    GAIN = auto()
    HOMDEL = auto()
    HETLOSS = auto()
    MRNA_HIGH = auto()
    MRNA_LOW = auto()
    PROTEIN_HIGH = auto()
    PROTEIN_LOW = auto()
    FUSION = auto()
    GERMLINE = auto()
    SPLICE = auto()
    SPLICE_PASSENGER = auto()
    MISSENSE = auto()
    MISSENSE_PASSENGER = auto()
    PROMOTER = auto()
    TRUNC = auto()
    TRUNC_PASSENGER = auto()
    INFRAME = auto()
    INFRAME_PASSENGER = auto()
    STRUCTURAL_VARIANT = auto()
    STRUCTURAL_VARIANT_PASSENGER = auto()
    # STRUCTURE_VARIANT

    OTHER = 10000


# Overwrite the default legend
# So that the legend entry will add a background
class _AltPiece:
    background_color = None

    def legend(self, x, y, w, h):
        arts = []
        if self.background_color is not None:
            arts.append(Rectangle((x, y), w, h, facecolor=self.background_color))
        arts.append(self.draw(x, y, w, h, None))
        return PatchCollection(arts, match_original=True)


class Rect(_AltPiece, mlayers.Rect):
    pass


class FrameRect(_AltPiece, mlayers.FrameRect):
    pass


class FracRect(_AltPiece, mlayers.FracRect):
    pass


@dataclass(repr=False)
class MatchRule:
    startswith: str = None
    endswith: str = None
    contains: List[str] = field(default_factory=list)
    flexible: bool = False

    def is_match(self, text: str):
        text = text.lower()

        match_start = not self.flexible
        if self.startswith is not None:
            match_start = text.startswith(self.startswith)

        match_end = not self.flexible
        if self.endswith is not None:
            match_end = text.endswith(self.endswith)

        match_contains = True
        if len(self.contains) > 0:
            match_contains = np.sum([i in text for i in self.contains])

        if self.flexible:
            return match_start | match_end | match_contains
        else:
            return match_start & match_end & match_contains


MATCH_POOL = {
    Alteration.AMP: MatchRule(startswith="amp"),
    Alteration.GAIN: MatchRule(startswith="gain"),
    Alteration.HOMDEL: MatchRule(
        startswith="homdel", contains=["deep", "deletion"], flexible=True
    ),
    Alteration.HETLOSS: MatchRule(
        startswith="hetloss", contains=["shallow", "deletion"], flexible=True
    ),
    Alteration.MRNA_HIGH: MatchRule(contains=["mrna", "high"]),
    Alteration.MRNA_LOW: MatchRule(contains=["mrna", "low"]),
    Alteration.PROTEIN_HIGH: MatchRule(contains=["protein", "high"]),
    Alteration.PROTEIN_LOW: MatchRule(contains=["protein", "low"]),
    Alteration.FUSION: MatchRule(startswith="fusion"),
    Alteration.GERMLINE: MatchRule(startswith="germline"),
    Alteration.MISSENSE_PASSENGER: MatchRule(
        startswith="missense", contains=["passenger"]
    ),
    Alteration.MISSENSE: MatchRule(
        startswith="missense", contains=["driver"], flexible=True
    ),
    Alteration.PROMOTER: MatchRule(startswith="promoter"),
    Alteration.TRUNC_PASSENGER: MatchRule(startswith="trunc", contains=["passenger"]),
    Alteration.TRUNC: MatchRule(startswith="trunc"),
    Alteration.INFRAME_PASSENGER: MatchRule(
        startswith="inframe", contains=["passenger"]
    ),
    Alteration.INFRAME: MatchRule(startswith="inframe"),
    Alteration.STRUCTURAL_VARIANT_PASSENGER: MatchRule(
        startswith="sv", contains=["structural", "variant", "passenger"], flexible=True
    ),
    Alteration.STRUCTURAL_VARIANT: MatchRule(
        startswith="sv", contains=["structural", "variant"], flexible=True
    ),
    Alteration.SPLICE_PASSENGER: MatchRule(startswith="splice", contains=["passenger"]),
    Alteration.SPLICE: MatchRule(startswith="splice"),
}

SHAPE_BANK = {
    Alteration.BACKGROUND: Rect(
        color=DEFAULT_GREY, label="No alterations", zorder=-10000
    ),
    # CNA
    Alteration.AMP: Rect(color=CNA_COLOR_AMP, label="Amplification"),
    # ShapeId.ampRectangle
    Alteration.GAIN: Rect(color=CNA_COLOR_GAIN, label="Gain"),
    # ShapeId.gainRectangle
    Alteration.HOMDEL: Rect(color=CNA_COLOR_HOMDEL, label="Deep Deletion"),
    Alteration.HETLOSS: Rect(color=CNA_COLOR_HETLOSS, label="Shallow Deletion"),
    # mRNA Regulation
    Alteration.MRNA_HIGH: FrameRect(
        color=MRNA_COLOR_HIGH, width=1.5, label="mRNA High", zorder=1000
    ),
    Alteration.MRNA_LOW: FrameRect(
        color=MRNA_COLOR_LOW, width=1.5, label="mRNA Low", zorder=1000
    ),
    # Protein Expression Regulation
    Alteration.PROTEIN_HIGH: FracRect(
        color=PROT_COLOR_HIGH, frac=(1.0, 0.2), label="Protein High", zorder=100
    ),
    Alteration.PROTEIN_LOW: FracRect(
        color=PROT_COLOR_LOW, frac=(1.0, 0.2), label="Protein Low", zorder=100
    ),
    # Structural variant
    Alteration.STRUCTURAL_VARIANT: FracRect(
        color=STRUCTURAL_VARIANT_COLOR,
        frac=(1.0, 0.6),
        label="Structural variant",
        zorder=200,
    ),
    Alteration.STRUCTURAL_VARIANT_PASSENGER: FracRect(
        color=STRUCTURAL_VARIANT_PASSENGER_COLOR,
        frac=(1.0, 0.6),
        label="Structural variant (putative passenger)",
        zorder=200,
    ),
    Alteration.FUSION: FracRect(
        color=MUT_COLOR_FUSION, frac=(1.0, 0.6), label="Fusion", zorder=200
    ),
    # Splice
    Alteration.SPLICE: FracRect(
        color=MUT_COLOR_SPLICE, frac=(1.0, 0.3), label="Splice Mutation", zorder=400
    ),
    Alteration.SPLICE_PASSENGER: FracRect(
        color=MUT_COLOR_SPLICE_PASSENGER,
        frac=(1.0, 0.3),
        label="Splice Mutation (putative passenger)",
        zorder=400,
    ),
    Alteration.MISSENSE: FracRect(
        color=MUT_COLOR_MISSENSE,
        frac=(1.0, 0.3),
        label="Mutation (putative driver)",
        zorder=400,
    ),
    Alteration.MISSENSE_PASSENGER: FracRect(
        color=MUT_COLOR_MISSENSE_PASSENGER,
        frac=(1.0, 0.3),
        label="Missense Mutation (putative passenger)",
        zorder=400,
    ),
    Alteration.OTHER: FracRect(
        color=MUT_COLOR_OTHER, frac=(1.0, 0.3), label="Other Mutation", zorder=400
    ),
    Alteration.PROMOTER: FracRect(
        color=MUT_COLOR_PROMOTER, frac=(1.0, 0.3), label="Promoter Mutation", zorder=400
    ),
    Alteration.TRUNC: FracRect(
        color=MUT_COLOR_TRUNC, frac=(1.0, 0.3), label="Truncating Mutation", zorder=400
    ),
    Alteration.TRUNC_PASSENGER: FracRect(
        color=MUT_COLOR_TRUNC_PASSENGER,
        frac=(1.0, 0.3),
        label="Truncating Mutation (putative passenger)",
        zorder=400,
    ),
    Alteration.INFRAME: FracRect(
        color=MUT_COLOR_INFRAME,
        frac=(1.0, 0.3),
        label="Inframe Mutation (putative driver)",
        zorder=400,
    ),
    Alteration.INFRAME_PASSENGER: FracRect(
        color=MUT_COLOR_INFRAME_PASSENGER,
        frac=(1.0, 0.3),
        label="Inframe Mutation (putative passenger)",
        zorder=400,
    ),
    # Germline
    Alteration.GERMLINE: FracRect(
        color=MUT_COLOR_GERMLINE, frac=(1.0, 0.1), label="Germline Mutation", zorder=600
    ),
}



from __future__ import annotations

from dataclasses import dataclass
from numbers import Number
from typing import List, Dict, Literal
from uuid import uuid4

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

from .exceptions import AppendLayoutError, DuplicateName
from .utils import _check_side


# If not specify, we have few default setting
# 1. Size is inch unit
# 2. Origin point is left-bottom

# Axes is a  rect, a rect can be recorded by a left-bottom anchor point
# and width and height. When other axes is added, the anchor point is aligned
# to either x-axis or y-axis, we could easily compute the final anchor point
# if we know the size of added axes

# It's unlikely to return axes right after being added, the split operation
# is unknown, this will create more than one axes.


def _split(chunk_ratios, spacing=0.05, group_ratios=None):
    """Return the relative anchor point, in ratio"""

    ratios = np.asarray(chunk_ratios) / np.sum(chunk_ratios)
    count = len(ratios)
    if isinstance(spacing, Number):
        spacing = [spacing for _ in range(count - 1)]
    spacing = np.asarray(spacing)

    canvas_size = 1 - np.sum(spacing)
    ratios = ratios * canvas_size
    assert np.sum(spacing) + np.sum(ratios) - 1 <= 0.00001

    if group_ratios is not None:
        c = np.asarray(group_ratios)
        groups = (c / c.sum() * count).astype(int)
        if np.sum(groups) != count:
            raise ValueError(f"Cannot group the split with ratios of {group_ratios}")

        regroup_ratios = []
        anchors = [0]

        start_anchor = 0
        start_ix = 0
        last_ix = len(groups) - 1
        for i, g in enumerate(groups):
            if g == 1:
                gratio = ratios[start_ix]
            else:
                mr = ratios[start_ix : start_ix + g].sum()
                ms = spacing[start_ix : start_ix + g - 1].sum()
                gratio = mr + ms

            if i < last_ix:
                gspacing = spacing[start_ix + g - 1]
                start_anchor += gratio + gspacing
                anchors.append(start_anchor)
            else:
                start_anchor += gratio
            regroup_ratios.append(gratio)
            start_ix += g
        ratios = regroup_ratios
    else:
        start = 0
        anchors = []
        last_ix = len(ratios) - 1
        for ix, i in enumerate(ratios):
            anchors.append(start)
            start += i
            # If not the last one
            # Add space
            if ix < last_ix:
                start += spacing[ix]
    return np.array(ratios), np.array(anchors)


def get_axes_rect(rect, figsize):
    cx, cy, cw, ch = rect
    fig_w, fig_h = figsize
    return cx / fig_w, cy / fig_h, cw / fig_w, ch / fig_h


class BaseCell:
    name: str
    side: str
    projection: str = None
    ax = None
    anchor = None

    is_split: bool = False
    h_anchors: np.ndarray = None
    h_ratios: np.ndarray = None
    v_anchors: np.ndarray = None
    v_ratios: np.ndarray = None

    is_canvas: bool = True

    def set_ax(self, ax):
        self.ax = ax

    def set_anchor(self, anchor):
        self.anchor = anchor

    def get_cell_size(self) -> (float, float):
        raise NotImplementedError

    def get_rect(self):
        cx, cy = self.anchor
        cw, ch = self.get_cell_size()

        return cx, cy, cw, ch

    def hsplit(self, chunk_ratios, spacing=0.05, group_ratios=None):
        """
        Parameters
        ----------
        chunk_ratios :
            The length of each chunk from top to bottom
        spacing :
            Relative to the cell size, the value should between 0~1
        group_ratios :
            Regroup the split chunks

        """
        ratios, anchors = _split(
            chunk_ratios[::-1], spacing=spacing, group_ratios=group_ratios
        )
        self.h_ratios = ratios
        self.h_anchors = anchors

    def vsplit(self, chunk_ratios, spacing=0.05, group_ratios=None):
        """
        Parameters
        ----------
        chunk_ratios :
            The length of each chunk from left to right
        spacing :
            Relative to the cell size, the value should between 0~1
        group_ratios :
            Regroup the split chunks

        """
        ratios, anchors = _split(
            chunk_ratios, spacing=spacing, group_ratios=group_ratios
        )
        self.v_ratios = ratios
        self.v_anchors = anchors

    def get_rects(self):
        cx, cy = self.anchor
        cw, ch = self.get_cell_size()

        if self.v_anchors is not None:
            w_anchors = self.v_anchors * cw + cx
            w_sizes = self.v_ratios * cw
        else:
            w_anchors = [cx]
            w_sizes = [cw]

        if self.h_anchors is not None:
            h_anchors = self.h_anchors * ch + cy
            h_sizes = self.h_ratios * ch
        else:
            h_anchors = [cy]
            h_sizes = [ch]

        # Make sure the order of split axes is
        # from top-left to bottom-right
        h_anchors = h_anchors[::-1]
        h_sizes = h_sizes[::-1]

        rects = []
        for ay, ay_size in zip(h_anchors, h_sizes):
            for ax, ax_size in zip(w_anchors, w_sizes):
                rects.append((ax, ay, ax_size, ay_size))
        return rects


@dataclass
class MainCell(BaseCell):
    name: str
    width: float
    height: float
    side: str = "main"
    is_canvas: bool = True
    projection: str = None

    def get_cell_size(self):
        return self.width, self.height


@dataclass
class GridCell(BaseCell):
    name: str
    side: str
    size: float
    attach: MainCell
    is_canvas: bool = True
    projection: str = None

    def get_cell_size(self):
        """Get width, height of a cell"""
        if self.side in ["top", "bottom"]:
            return self.attach.width, self.size
        else:
            return self.size, self.attach.height


@dataclass
class Margin:
    top: float
    right: float
    bottom: float
    left: float


class _MarginMixin:
    margin = Margin(0, 0, 0, 0)

    def set_margin(self, margin):
        if isinstance(margin, Number):
            self.margin = Margin(margin, margin, margin, margin)
        else:
            if len(margin) == 4:
                self.margin = Margin(*margin)
            else:
                msg = (
                    "margin must be one number or a tuple with 4 numbers"
                    "(top, right, bottom, left)"
                )
                raise ValueError(msg)

    def get_margin_w(self):
        return self.margin.left + self.margin.right

    def get_margin_h(self):
        return self.margin.top + self.margin.bottom


class CrossLayout(_MarginMixin):
    """The cross-layout engine

    This class implements the cross-layout. The axes are added
    incrementally to the main axes.

    - The size is always in the unit of inches


    Parameters
    ----------
    name : str
        The name of the main canvas
    width : float
        The width of the main canvas
    height : float
        The height of the main canvas
    init_main : bool
        Add the main canvas as axes
    projection : str
        The projection of the main canvas

    Attributes
    ----------
    is_composite : bool
        Use to indicate the current position of CrossLayout.
        Whether it's in a CompositeCrossLayout.

    """

    def __init__(
        self, name, width, height, init_main=True, projection=None, margin=0.2
    ):
        self._legend_ax_name = None
        self.main_cell = MainCell(
            name, width, height, is_canvas=init_main, projection=projection
        )
        self._side_cells = {"top": [], "bottom": [], "left": [], "right": []}
        self.cells: Dict[str, BaseCell] = {name: self.main_cell}
        self._pads = {}

        self.is_composite = False
        self.figure = None
        self.figsize = None
        # The left bottom anchor point of the layout
        self.anchor = None
        # (top, right, bottom, left)
        self.set_margin(margin)

    def _get_cell(self, name):
        cell = self.cells.get(name)
        if cell is None:
            raise ValueError(f"Axes with name {name} not exist")
        return cell

    def add_ax(self, side, name, size, pad=0.0, projection=None):
        """Add an axes to the layout

        Parameters
        ----------
        side
        name
        size
        pad
        projection

        Returns
        -------

        """
        _check_side(side)
        if self.cells.get(name) is not None:
            raise DuplicateName(name)

        new_cell = GridCell(
            name=name,
            side=side,
            size=size,
            attach=self.main_cell,
            projection=projection,
        )
        # Add pad before add canvas
        if pad > 0.0:
            self.add_pad(side, pad)
        self._side_cells[side].append(new_cell)
        self.cells[name] = new_cell

    def add_pad(self, side, size):
        """Add padding between axes

        Parameters
        ----------
        side
        size

        Returns
        -------

        """
        _check_side(side)
        new_pad = GridCell(
            name=uuid4().hex,
            side=side,
            size=size,
            is_canvas=False,
            attach=self.main_cell,
        )
        self._side_cells[side].append(new_pad)

    def remove_ax(self, name):
        cell = self.cells.pop(name, None)
        if cell is not None:
            # remove from self._side_cells
            side = cell.side
            self._side_cells[side].remove(cell)
            # remove its pad
            pad = self._pads.pop(name, None)
            if pad is not None:
                self._side_cells[side].remove(pad)

    def get_ax(self, name):
        return self.cells[name].ax

    def get_main_ax(self):
        return self.main_cell.ax

    def add_legend_ax(self, side, size, pad=0.0):
        """Add a legend axes

        Parameters
        ----------
        side
        size
        pad

        Returns
        -------

        """
        # TODO: Ensure the legend axes always add at last
        self._legend_ax_name = "-".join([self.main_cell.name, "legend", uuid4().hex])
        self.add_ax(side, self._legend_ax_name, size, pad=pad)

    def get_legend_ax(self):
        if self._legend_ax_name is not None:
            return self.get_ax(self._legend_ax_name)

    def remove_legend_ax(self):
        self.remove_ax(self._legend_ax_name)

    def set_legend_size(self, size):
        legend_cell = self._get_cell(self._legend_ax_name)
        legend_cell.size = size

    def vsplit(self, name, chunk_ratios, spacing=0.05, group_ratios=None):
        cell = self._get_cell(name)
        cell.is_split = True
        cell.vsplit(chunk_ratios, spacing=spacing, group_ratios=group_ratios)

    def hsplit(self, name, chunk_ratios, spacing=0.05, group_ratios=None):
        cell = self._get_cell(name)
        cell.is_split = True
        cell.hsplit(chunk_ratios, spacing=spacing, group_ratios=group_ratios)

    def is_split(self, name):
        """Query if a cell is split"""
        return self._get_cell(name).is_split

    def get_side_size(self, side):
        return np.sum([c.size for c in self._side_cells[side]])

    def get_bbox_width(self):
        """Get the bbox width in inches"""
        box_w = (
            self.main_cell.width
            + self.get_side_size("left")
            + self.get_side_size("right")
        )
        return box_w

    def get_bbox_height(self):
        """Get the bbox height in inches"""
        box_h = (
            self.main_cell.height
            + self.get_side_size("top")
            + self.get_side_size("bottom")
        )
        return box_h

    def get_bbox_size(self):
        """Get the bbox size in inches (width, height)"""
        return self.get_bbox_width(), self.get_bbox_height()

    def get_figure_size(self):
        """Get the figsize in inches (width, height)"""
        if self.anchor is None:
            if self.is_composite:
                return self.get_bbox_size()
            w, h = self.get_bbox_size()
            return w + self.get_margin_w(), h + self.get_margin_h()

        else:
            ox, oy = self.anchor
            fig_w = ox + self.get_main_width() + self.get_side_size("right")
            fig_h = ox + self.get_main_height() + self.get_side_size("top")
            if self.is_composite:
                return fig_w, fig_h
            return fig_w + self.get_margin_w(), fig_h + self.get_margin_h()

    def get_main_anchor(self):
        """Get the main anchor point"""
        if self.anchor is None:
            x = self.get_side_size("left")
            y = self.get_side_size("bottom")
            if self.is_composite:
                return x, y
            x += self.margin.bottom
            y += self.margin.left
            return x, y
        else:
            return self.anchor

    def set_layout(self, main_anchor):
        """Set the layout of all cell

        Parameters
        ----------
        main_anchor : point
            The anchor point of the main cell

        """
        self.main_cell.set_anchor(main_anchor)
        ax, ay = main_anchor
        for side, cells in self._side_cells.items():
            if side == "top":
                cell_y = ay + self.main_cell.height
                for c in cells:
                    c.set_anchor((ax, cell_y))
                    cell_y += c.size
            elif side == "bottom":
                cell_y = ay
                for c in cells:
                    cell_y -= c.size
                    c.set_anchor((ax, cell_y))
            elif side == "left":
                cell_x = ax
                for c in cells:
                    cell_x -= c.size
                    c.set_anchor((cell_x, ay))
            else:
                cell_x = ax + self.main_cell.width
                for c in cells:
                    c.set_anchor((cell_x, ay))
                    cell_x += c.size

    def get_main_height(self):
        return self.main_cell.height

    def set_main_height(self, height):
        self.main_cell.height = height

    def get_main_width(self):
        return self.main_cell.width

    def set_main_width(self, width):
        self.main_cell.width = width

    def set_anchor(self, anchor):
        self.anchor = anchor

    def set_bbox_anchor(self, anchor):
        xoff = self.margin.left + self.get_side_size("left")
        yoff = self.margin.bottom + self.get_side_size("bottom")
        self.anchor = anchor[0] + xoff, anchor[1] + yoff

    def set_figsize(self, figsize):
        self.figsize = figsize

    def set_render_size(self, name, size):
        self._get_cell(name).size = size

    def initiate_axes(self, figure, _debug=False):
        figsize = self.figsize
        clear_axes = figure == self.figure
        # add axes
        for c in list(self.cells.values()):
            # Previous axes will be removed
            if clear_axes:
                _remove_axes(c.ax)
            # only add if it's a canvas
            if c.is_canvas:
                # check if it's split
                if c.is_split:
                    ax_rects = c.get_rects()
                    ax_rects = [get_axes_rect(r, figsize) for r in ax_rects]
                    axes = []
                    for ix, rect in enumerate(ax_rects):
                        ax = figure.add_axes(rect, projection=c.projection)
                        axes.append(ax)
                        if _debug:
                            _debug_ax(ax, side=c.side, text=ix + 1)

                    c.set_ax(axes)
                else:
                    ax_rect = c.get_rect()
                    ax_rect = get_axes_rect(ax_rect, figsize)
                    ax = figure.add_axes(ax_rect, projection=c.projection)
                    c.set_ax(ax)
                    if _debug:
                        _debug_ax(ax, side=c.side, text=f"{c.name}{c.get_cell_size()}")

    def freeze(
        self,
        figure=None,
        scale=1,
        _debug=False,
    ):
        """Freeze the current layout and draw on figure

        Freeze is safe to be called multiple times, it will update
        the layout if draw on the same figure

        Parameters
        ----------
        figure : :class:`matplotlib.figure.Figure`
            A matplotlib figure instance
        scale : float
            To scale the size of figure
        _debug : bool
            Not a public parameter, for internal debug use,
            the information display on the canvas
            is not guarantee to be consistent.

        Returns
        -------

        """
        # If not composed, update the figsize
        if not self.is_composite:
            self.figsize = np.array(self.get_figure_size())
        figsize = self.figsize * scale
        if figure is None:
            figure = plt.figure(figsize=figsize)
        else:
            if not self.is_composite:
                figure.set_size_inches(*figsize)

        main_anchor = self.get_main_anchor()
        self.set_layout(main_anchor)

        self.initiate_axes(figure, _debug=_debug)
        self.figure = figure

    def plot(self, scale=1):
        self.freeze(scale=scale, _debug=True)

    def _append_check(self, other):
        if isinstance(other, CompositeCrossLayout):
            raise AppendLayoutError
        if isinstance(other, (CrossLayout, Number)):
            c = CompositeCrossLayout(main_layout=self)
            return c

    def append(self, side, other):
        c = self._append_check(other)
        c.append(side, other)

    def __truediv__(self, other):
        return self.append("bottom", other)

    def __add__(self, other):
        return self.append("right", other)


def _debug_ax(ax, side, text=None):
    close_ticks(ax)
    options = dict(transform=ax.transAxes, va="center", ha="center")
    if side == "left":
        options["rotation"] = 90
    elif side == "right":
        options["rotation"] = -90

    if text is not None:
        ax.text(0.5, 0.5, text, **options)


def _remove_axes(ax: Axes | List[Axes]):
    if ax is None:
        return
    if isinstance(ax, Axes):
        ax.remove()
    else:
        for a in ax:
            a.remove()


def close_ticks(ax):
    ax.tick_params(
        bottom=False,
        top=False,
        left=False,
        right=False,
        labelbottom=False,
        labeltop=False,
        labelleft=False,
        labelright=False,
    )


def _reset_layout(layout: CrossLayout):
    layout.set_anchor((0, 0))
    layout.is_composite = True
    return layout


@dataclass
class _LegendAxes:
    side: str
    size: float
    pad: float
    ax = None

    def get_length(self):
        return self.size + self.pad


class CompositeCrossLayout(_MarginMixin):
    """A class to layout multiple Cross Layouts

    .. warning::
        This class are not supposed to be used directly by user

    Parameters
    ----------
    main_layout : CrossLayout
        The main cross layout

    """

    figure = None

    def __init__(self, main_layout, margin=0, align_main=True) -> None:
        self.main_layout = _reset_layout(main_layout)
        self.main_cell_height = self.main_layout.get_main_height()
        self.main_cell_width = self.main_layout.get_main_width()
        self._side_layouts: Dict[str, List[CrossLayout]] = {
            "top": [],
            "bottom": [],
            "right": [],
            "left": [],
        }
        self._legend_axes = None
        self.layouts = {self.main_layout.main_cell.name: self.main_layout}
        self.set_margin(margin)
        self.align_main = align_main

    def append(self, side, other):
        _check_side(side)
        if isinstance(other, CompositeCrossLayout):
            raise AppendLayoutError

        elif isinstance(other, Number):
            if side in ["left", "right"]:
                width, height = other, self.main_cell_height
            else:
                width, height = self.main_cell_height, other
            other = CrossLayout(
                name=uuid4().hex, width=width, height=height, init_main=False
            )
            other.is_composite = True
            self._side_layouts[side].append(other)
        elif isinstance(other, CrossLayout):
            other = _reset_layout(other)
            if self.align_main:
                adjust = "height" if side in ["left", "right"] else "width"
                adjust_size = getattr(self, f"main_cell_{adjust}")
                getattr(other, f"set_main_{adjust}").__call__(adjust_size)
            self._side_layouts[side].append(other)
            self.layouts[other.main_cell.name] = other
        else:
            raise TypeError(f"Cannot append object type of {type(other)}")

    def __truediv__(self, other: CrossLayout):
        self.append("bottom", other)

    def __add__(self, other: CrossLayout):
        self.append("right", other)

    def add_legend_ax(self, side, size, pad=0.0):
        """Extend the layout

        This is used to draw legends after concatenation

        """
        self._legend_axes = _LegendAxes(side=side, size=size, pad=pad)

    def get_legend_ax(self):
        if self._legend_axes is not None:
            return self._legend_axes.ax

    def set_legend_size(self, size):
        self._legend_axes.size = size

    def get_side_size(self, side):
        """Get the max extend of a side, the center is the
        main cell of the main layout, not the whole main layout
        legend axes included"""
        # It's possible the left side of top cross-layout
        # is longer than the left cross-layout

        size = 0
        other_size = []
        size += self.main_layout.get_side_size(side)
        if side in ["left", "right"]:
            for g in self._side_layouts.get(side):
                size += g.get_bbox_width()

            for g in self._side_layouts["top"] + self._side_layouts["bottom"]:
                other_size.append(g.get_side_size(side))
        else:
            for g in self._side_layouts.get(side):
                size += g.get_bbox_height()

            for g in self._side_layouts["left"] + self._side_layouts["right"]:
                other_size.append(g.get_side_size(side))

        other_size = np.max(other_size, initial=0)
        legend_size = 0
        if self._legend_axes is not None:
            if self._legend_axes.side == side:
                legend_size = self._legend_axes.get_length()
        return max(size, other_size) + legend_size

    def get_bbox_width(self):
        return (
            self.main_cell_width
            + self.get_side_size("left")
            + self.get_side_size("right")
        )

    def get_bbox_height(self):
        return (
            self.main_cell_height
            + self.get_side_size("top")
            + self.get_side_size("bottom")
        )

    def get_bbox_size(self):
        """Get the minimum figsize that fits all layouts inside"""
        return self.get_bbox_width(), self.get_bbox_height()

    def get_figure_size(self):
        """Figure size is bbox size + margin"""
        fig_w = self.get_bbox_width() + self.get_margin_w()
        fig_h = self.get_bbox_height() + self.get_margin_h()
        return fig_w, fig_h

    def get_main_anchor(self):
        x = self.get_side_size("left") + self.margin.left
        y = self.get_side_size("bottom") + self.margin.bottom
        return x, y

    def set_anchor(self, anchor):
        self.main_layout.set_anchor(anchor)

    def freeze(self, figure=None, scale=1, _debug=False):
        figsize = np.asarray(self.get_figure_size()) * scale

        if figure is None:
            figure = plt.figure(figsize=figsize)
        else:
            figure.set_size_inches(figsize)

        # To freeze all the layouts
        # We only need to compute the anchor point for main layouts
        # And then the location will be automatically derived

        # compute the anchor point for the main layout
        mx, my = self.get_main_anchor()
        self.main_layout.set_anchor((mx, my))
        self.main_layout.set_figsize(figsize)

        self.main_layout.freeze(figure, _debug=_debug)

        # The left and right share the y
        offset_x = mx - self.main_layout.get_side_size("left")
        for g in self._side_layouts["left"]:
            offset_x -= g.get_side_size("right") + g.get_main_width()
            g.set_anchor((offset_x, my))
            g.set_figsize(figsize)
            g.freeze(figure, _debug=_debug)
            offset_x -= g.get_side_size("left")

        offset_x = mx + self.main_cell_width + self.main_layout.get_side_size("right")
        for g in self._side_layouts["right"]:
            offset_x += g.get_side_size("left")
            g.set_anchor((offset_x, my))
            g.set_figsize(figsize)
            g.freeze(figure, _debug=_debug)
            offset_x += g.get_main_width() + g.get_side_size("right")

        # The top and bottom share the x
        offset_y = my - self.main_layout.get_side_size("bottom")
        for g in self._side_layouts["bottom"]:
            offset_y -= g.get_side_size("top") + g.get_main_height()
            g.set_anchor((mx, offset_y))
            g.set_figsize(figsize)
            g.freeze(figure, _debug=_debug)
            offset_y -= g.get_side_size("bottom")

        offset_y = my + self.main_cell_height + self.main_layout.get_side_size("top")
        for g in self._side_layouts["top"]:
            offset_y += g.get_side_size("bottom")
            g.set_anchor((mx, offset_y))
            g.set_figsize(figsize)
            g.freeze(figure, _debug=_debug)
            offset_y += g.get_main_height() + g.get_side_size("top")

        # Add legend axes
        # The range is limited to the main_cell in each layout
        if self._legend_axes is not None:
            xmin, ymin, xmax, ymax = self._edge_main_point()
            legend_w, legend_h = xmax - xmin, ymax - ymin
            fig_w, fig_h = figsize

            side = self._legend_axes.side
            size = self._legend_axes.size

            if side == "right":
                cx, cy = fig_w - size - self.margin.right, ymin
                cw, ch = size, legend_h
            elif side == "left":
                cx, cy = self.margin.left, ymin
                cw, ch = size, legend_h
            elif side == "top":
                cx, cy = xmin, fig_h - self.margin.top - size
                cw, ch = legend_w, size
            elif side == "bottom":
                cx, cy = xmin, self.margin.bottom
                cw, ch = legend_w, size

            rect = get_axes_rect((cx, cy, cw, ch), figsize)
            ax = figure.add_axes(rect)
            if _debug:
                _debug_ax(ax, side=side, text="Legend Axes")
            self._legend_axes.ax = ax

        self.figure = figure

    def _edge_main_point(self):
        """This will return the edge point of all the main cell"""
        left_side = self._side_layouts["left"]
        if len(left_side) > 0:
            xmin = left_side[-1].main_cell.anchor[0]
        else:
            xmin = self.main_layout.main_cell.anchor[0]

        right_side = self._side_layouts["right"]
        if len(right_side) > 0:
            cell = right_side[-1].main_cell
        else:
            cell = self.main_layout.main_cell
        xmax = cell.anchor[0] + cell.width

        bottom_side = self._side_layouts["bottom"]
        if len(bottom_side) > 0:
            ymin = bottom_side[-1].main_cell.anchor[1]
        else:
            ymin = self.main_layout.main_cell.anchor[1]

        top_side = self._side_layouts["top"]
        if len(top_side) > 0:
            cell = top_side[-1].main_cell
        else:
            cell = self.main_layout.main_cell
        ymax = cell.anchor[1] + cell.height

        return xmin, ymin, xmax, ymax

    def plot(self, scale=1):
        self.freeze(scale=scale, _debug=True)

    def get_ax(self, layout_name, ax_name):
        return self.layouts[layout_name].get_ax(ax_name)

    def get_main_ax(self, layout_name):
        return self.layouts[layout_name].get_main_ax()


class StackCrossLayout(_MarginMixin):
    """A stack of cross layouts

    This class allow users to stack multiple cross layouts
    either horizontally or vertically

    Multiple StackCrossLayout can also be stacked

    .. warning::
        This class are not supposed to be used directly by user

    Parameters
    ----------
    layouts : list of :class:`CrossLayout`, :class:`StackCrossLayout`
        The layouts to be stacked
    direction : {"horizontal", "vertical"}
        The direction of the stack, horizontal will stack from left to right
        vertical will stack from top to bottom
    align : {"center", "bottom", "top", "left", "right"}
        The alignment of the stack, the default is center

    """

    _direction_align = {
        "horizontal": {"center", "top", "bottom"},
        "vertical": {"center", "left", "right"},
    }

    def __init__(
        self,
        layouts: List[CrossLayout | StackCrossLayout],
        direction="horizontal",
        align: Literal["center", "bottom", "top", "left", "right"] = "center",
        spacing=0,
        margin=0,
        name=None,
    ):
        # Check the direction and align
        if direction not in self._direction_align:
            raise ValueError(f"Invalid direction {direction}")
        if align not in self._direction_align[direction]:
            raise ValueError(
                f"When setting direction={direction}, "
                f"align must be one of {self._direction_align[direction]}"
            )

        self.layouts = []
        self._layouts_mapper = {}
        for layout in layouts:
            layout = _reset_layout(layout)
            self.layouts.append(layout)
            if hasattr(layout, "name"):
                self._layouts_mapper[layout.name] = layout
            else:
                self._layouts_mapper[layout.main_cell.name] = layout
        self.direction = direction
        self.align = align
        self.spacing = spacing
        if name is None:
            name = uuid4().hex
        self.name = name
        self.set_margin(margin)
        self.is_composite = False
        self.figure = None
        self.figsize = None
        # The left bottom anchor point of the layout
        self.anchor = None
        self._legend_axes = None

    def remove_legend_ax(self):
        self._legend_axes = None
        for layout in self.layouts:
            layout.remove_legend_ax()

    def _get_layout_widths(self):
        return [layout.get_bbox_width() for layout in self.layouts]

    def _get_layout_heights(self):
        return [layout.get_bbox_height() for layout in self.layouts]

    def _get_spacing_widths(self):
        return self.spacing * (len(self.layouts) - 1)

    def _get_spacing_heights(self):
        return self.spacing * (len(self.layouts) - 1)

    def get_bbox_width(self):
        ws = self._get_layout_widths()
        if self.direction == "horizontal":
            return np.sum(ws) + self._get_spacing_widths()
        else:
            if self.align == "center":
                return np.max(ws)
            elif self.align == "left":
                left_sides = np.asarray(
                    [layout.get_side_size("left") for layout in self.layouts]
                )
                right_leftover = ws - left_sides
                return np.max(left_sides) + np.max(right_leftover)
            else:
                right_sides = np.asarray(
                    [layout.get_side_size("right") for layout in self.layouts]
                )
                left_leftover = ws - right_sides
                return np.max(right_sides) + np.max(left_leftover)

    def get_bbox_height(self):
        hs = self._get_layout_heights()
        if self.direction == "vertical":
            return np.sum(hs) + self._get_spacing_heights()
        else:
            if self.align == "center":
                return np.max(hs)
            elif self.align == "top":
                top_sides = np.asarray(
                    [layout.get_side_size("top") for layout in self.layouts]
                )
                bottom_leftover = hs - top_sides
                return np.max(top_sides) + np.max(bottom_leftover)
            else:
                bottom_sides = np.asarray(
                    [layout.get_side_size("bottom") for layout in self.layouts]
                )
                top_leftover = hs - bottom_sides
                return np.max(bottom_sides) + np.max(top_leftover)

    def get_bbox_size(self):
        return self.get_bbox_width(), self.get_bbox_height()

    def get_figure_size(self):
        fig_w = self.get_bbox_width() + self.get_margin_w()
        fig_h = self.get_bbox_height() + self.get_margin_h()
        return fig_w, fig_h

    def set_figsize(self, figsize):
        self.figsize = figsize
        for layout in self.layouts:
            layout.set_figsize(figsize)

    # Mimic the CrossLayoutAPI
    def get_side_size(self, side):
        legend_size = 0
        if self._legend_axes is not None:
            if self._legend_axes.side == side:
                legend_size = self._legend_axes.get_length()
        return self._get_layouts_offset(side) + legend_size

    # Mimic the CrossLayoutAPI
    def get_main_height(self):
        return (
            self.get_bbox_height()
            - self._get_layouts_offset("bottom")
            - self._get_layouts_offset("top")
        )

    # Mimic the CrossLayoutAPI
    def get_main_width(self):
        return (
            self.get_bbox_width()
            - self._get_layouts_offset("left")
            - self._get_layouts_offset("right")
        )

    def _get_layouts_offset(self, side):
        if self.direction == "horizontal":
            if side in {"bottom", "top"}:
                return np.max([layout.get_side_size(side) for layout in self.layouts])
            elif side == "left":
                return self.layouts[0].get_side_size("left")
            else:
                return self.layouts[-1].get_side_size("right")
        else:
            if side in {"left", "right"}:
                return np.max([layout.get_side_size(side) for layout in self.layouts])
            elif side == "bottom":
                return self.layouts[-1].get_side_size("bottom")
            else:
                return self.layouts[0].get_side_size("top")

    def get_layout_anchors(self):
        """Get the anchor points of all layouts assume the layout is not composite"""
        base_x, base_y = self.margin.left, self.margin.bottom
        xs, ys = [], []
        if self.direction == "horizontal":
            # x is always the same regardless of the alignment
            for layout in self.layouts:
                base_x += layout.get_side_size("left")
                xs.append(base_x)
                base_x += (
                    layout.get_main_width()
                    + layout.get_side_size("right")
                    + self.spacing
                )
            # only y is different
            if self.align == "bottom":
                base_y = self.margin.bottom + self._get_layouts_offset("bottom")
                ys = [base_y for _ in range(len(self.layouts))]
            elif self.align == "top":
                base_y = (
                    self.margin.bottom
                    + self.get_bbox_height()
                    - self._get_layouts_offset("top")
                )
                ys = [base_y - layout.get_main_height() for layout in self.layouts]
            else:
                center_y = self.margin.bottom + self.get_bbox_height() / 2
                ys = [
                    center_y - layout.get_main_height() / 2 for layout in self.layouts
                ]
        else:
            # y is always the same regardless of the alignment
            for layout in self.layouts[::-1]:
                base_y += layout.get_side_size("bottom")
                ys.append(base_y)
                base_y += (
                    layout.get_main_height()
                    + layout.get_side_size("top")
                    + self.spacing
                )
            # only x is different
            if self.align == "left":
                base_x = self.margin.left + self._get_layouts_offset("left")
                xs = [base_x for _ in range(len(self.layouts))]
            elif self.align == "right":
                base_x = (
                    self.margin.left
                    + self.get_bbox_width()
                    - self._get_layouts_offset("right")
                )
                xs = [base_x - layout.get_main_width() for layout in self.layouts]
            else:
                center_x = self.margin.left + self.get_bbox_width() / 2
                xs = [center_x - layout.get_main_width() / 2 for layout in self.layouts]
            ys = ys[::-1]

        return np.array(list(zip(xs, ys)))

    def set_layout_anchors(self, anchors):
        for layout, a in zip(self.layouts, anchors):
            layout.set_anchor(a)

    def set_anchor(self, anchor):
        self.anchor = anchor

    def add_legend_ax(self, side, size, pad=0.0):
        """Extend the layout

        This is used to draw legends after concatenation

        """
        self._legend_axes = _LegendAxes(side=side, size=size, pad=pad)

    def get_legend_ax(self):
        if self._legend_axes is not None:
            return self._legend_axes.ax

    def set_legend_size(self, size):
        self._legend_axes.size = size

    def freeze(self, figure=None, scale=1, _debug=False):
        # If not composed, update the figsize
        if not self.is_composite:
            self.figsize = np.array(self.get_figure_size())
        figsize = self.figsize * scale
        if figure is None:
            figure = plt.figure(figsize=figsize)
        else:
            if not self.is_composite:
                figure.set_size_inches(*figsize)

        # Compute the anchor points for all sub layouts
        anchors = self.get_layout_anchors()
        # Offset the anchor point by the main anchor
        main_anchor = np.min(anchors, axis=0)
        if self.anchor is not None:
            xoff = self.anchor[0] - main_anchor[0]
            yoff = self.anchor[1] - main_anchor[1]
            anchors = [(x + xoff, y + yoff) for x, y in anchors]
        self.set_layout_anchors(anchors)

        for layout in self.layouts:
            layout.set_figsize(figsize)
            layout.freeze(figure, _debug=_debug)

        if self._legend_axes is not None:
            bbox_w, bbox_h = self.get_bbox_size()
            ax, ay = main_anchor

            side = self._legend_axes.side
            size = self._legend_axes.size

            if side == "right":
                cx, cy = ax + bbox_w, ay
                cw, ch = size, bbox_h
            elif side == "left":
                cx, cy = ax - size, ay
                cw, ch = size, bbox_h
            elif side == "top":
                cx, cy = ax, ay + bbox_h
                cw, ch = bbox_w, size
            elif side == "bottom":
                cx, cy = ax, ay - size
                cw, ch = bbox_w, size

            rect = get_axes_rect((cx, cy, cw, ch), figsize)
            ax = figure.add_axes(rect)
            if _debug:
                _debug_ax(ax, side=side, text="Legend Axes")
            self._legend_axes.ax = ax

        self.figure = figure

    def get_ax(self, layout_name, ax_name):
        return self._layouts_mapper[layout_name].get_ax(ax_name)

    def get_main_ax(self, layout_name):
        return self._layouts_mapper[layout_name].get_main_ax()
